//-------------------------------------------------------------------------!
//                                                                         !
//        N  A  S     P A R A L L E L     B E N C H M A R K S  3.3         !
//                                                                         !
//                                   S P                                   !
//                                                                         !
//-------------------------------------------------------------------------!
//                                                                         !
//    This benchmark is part of the NAS Parallel Benchmark 3.3 suite.      !
//    It is described in NAS Technical Reports 95-020 and 02-007           !
//                                                                         !
//    Permission to use, copy, distribute and modify this software         !
//    for any purpose with or without fee is hereby granted.  We           !
//    request, however, that all derived work reference the NAS            !
//    Parallel Benchmarks 3.3. This software is provided "as is"           !
//    without express or implied warranty.                                 !
//                                                                         !
//    Information on NPB 3.3, including the technical report, the          !
//    original specifications, source code, results and information        !
//    on how to submit new results, is available at:                       !
//                                                                         !
//           http://www.nas.nasa.gov/Software/NPB/                         !
//                                                                         !
//    Send comments or suggestions to  npb@nas.nasa.gov                    !
//                                                                         !
//          NAS Parallel Benchmarks Group                                  !
//          NASA Ames Research Center                                      !
//          Mail Stop: T27A-1                                              !
//          Moffett Field, CA   94035-1000                                 !
//                                                                         !
//          E-mail:  npb@nas.nasa.gov                                      !
//          Fax:     (650) 604-3957                                        !
//                                                                         !
//-------------------------------------------------------------------------!


//---------------------------------------------------------------------
//
// Authors: R. F. Van der Wijngaart
//          W. Saphir
//          H. Jin  (UPC version)
//---------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#define G_MAIN
#include "header.h"
#include "upcnpb.h"

#define BSIZE 132

shared [] int sh_ibuf[5];
shared [] double sh_rbuf[2];

//---------------------------------------------------------------------
//      program MPSP
//---------------------------------------------------------------------
int main(int argc, char *argv[]) {

       int    i, j, niter, step, c, error, fstatus;
       double mflops, t, tmax;
       int    verified;
       char   class;

       FILE *fp;
       char cbuf[BSIZE];

#ifdef USE_MTIMERS
       double t1[T_LAST+2], tsum[T_LAST+2], tming[T_LAST+2], tmaxg[T_LAST+2];
       char   t_recs[T_LAST+2][9];
#endif

       setup_upc(&argc, &argv);
       if (! active) goto done;

//---------------------------------------------------------------------
//      Root node reads input file (if it exists) else takes
//      defaults from parameters
//---------------------------------------------------------------------
       if (node == root) {
          
          printf("\n\n NAS Parallel Benchmarks 3.3 -- SP Benchmark\n");

          fp = fopen("timer.flag", "r");
          timeron = 0;
          if (fp) {
             timeron = 1;
             fclose(fp);
          }

          fp = fopen("inputsp.data", "r");
//
          if (fp != NULL) {
            printf(" Reading from input file inputsp.data\n");
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%d", &niter);
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%lf", &dt);
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%d %d %d", 
                   &grid_points(1), &grid_points(2), &grid_points(3));
            fclose(fp);
          } else {
            printf(" No input file inputsp.data. Using compiled defaults\n");
            niter = NITER_DEFAULT;
            dt    = dt_default;
            grid_points(1) = PROBLEM_SIZE;
            grid_points(2) = PROBLEM_SIZE;
            grid_points(3) = PROBLEM_SIZE;
          }

          printf(" Size: %4dx%4dx%4d\n", 
                 grid_points(1), grid_points(2), grid_points(3));
          printf(" Iterations: %4d    dt: %11.7f\n", niter, dt);
          if (no_nodes != total_nodes)
              printf(" Total number of processes: %5d\n", total_nodes);
          if (no_nodes != MAXCELLS*MAXCELLS) 
              printf(" WARNING: compiled for %5d processes\n",
                     MAXCELLS*MAXCELLS);
          printf(" Number of active processes: %5d\n\n", no_nodes);

       }

       if (node == root && no_nodes > 1) {
          sh_ibuf[0] = niter;
          sh_ibuf[1] = grid_points(1);
          sh_ibuf[2] = grid_points(2);
          sh_ibuf[3] = grid_points(3);
          sh_ibuf[4] = timeron;
          sh_rbuf[0] = dt;
       }
       upc_barrier;
       if (node != root) {
          niter          = sh_ibuf[0];
          grid_points(1) = sh_ibuf[1];
          grid_points(2) = sh_ibuf[2];
          grid_points(3) = sh_ibuf[3];
          timeron        = sh_ibuf[4];
          dt             = sh_rbuf[0];
       }


       make_set();

       for (c = 1; c <= ncells; c++) {
          if ( (cell_size(1,c) > IMAX) ||
               (cell_size(2,c) > JMAX) ||
               (cell_size(3,c) > KMAX) ) {
             printf(" %d %d %d %d %d\n", node, c, cell_size(1,c),
                     cell_size(2,c), cell_size(3,c));
             printf(" Problem size too big for compiled array sizes\n");
             goto done;
          }
       }

       for (i = 1; i <= T_LAST; i++)
          timer_clear(i);

       set_constants();

       initialize();

       lhsinit();

       exact_rhs();

       compute_buffer_size(5);

//---------------------------------------------------------------------
//      do one time step to touch all code, and reinitialize
//---------------------------------------------------------------------
       adi();
       initialize();

//---------------------------------------------------------------------
//      Synchronize before placing time stamp
//---------------------------------------------------------------------
       for (i = 1; i <= T_LAST; i++)
          timer_clear(i);
       upc_barrier;

       timer_start(1);

       for (step = 1; step <= niter; step++) {

          if (node == root) {
             if ((step%20) == 0 || 
                 step == 1) {
                printf(" Time step %4d\n", step);
             }
          }

          adi();

       }

       timer_stop(1);
       t = timer_read(1);
       
       verify(niter, &class, &verified);

       mupc_reduce_r8(&t, &tmax, 1, SMP_OPR_MAX, root);

       if( node == root ) {
          if( tmax != 0. ) {
             mflops = (881.174*SQR( PROBLEM_SIZE )*PROBLEM_SIZE
                      -4683.91*SQR( PROBLEM_SIZE )
                      +11484.5*(double)( PROBLEM_SIZE )
                      -19272.4) * (double)( niter ) / (tmax*1000000.0e0);
          } else {
             mflops = 0.0;
          }

         c_print_results("SP", class, grid_points(1), 
           grid_points(2), grid_points(3), niter, /*MAXCELLS*MAXCELLS, */
           total_nodes, tmax, mflops, "          floating point", 
           verified, NPBVERSION,COMPILETIME, NPB_CS1, NPB_CS2, NPB_CS3, NPB_CS4, NPB_CS5, 
           NPB_CS6, "(none)");
       }

#ifdef USE_MTIMERS
       if (!timeron) goto done;

       for (i = 1; i <= T_LAST; i++)
          t1[i] = timer_read(i);

       t1[T_XSOLVE] = t1[T_XSOLVE] - t1[T_XCOMM];
       t1[T_YSOLVE] = t1[T_YSOLVE] - t1[T_YCOMM];
       t1[T_ZSOLVE] = t1[T_ZSOLVE] - t1[T_ZCOMM];
       t1[0] = t1[T_XCOMM]+t1[T_YCOMM]+t1[T_ZCOMM]+t1[T_EXCH];
       t1[T_LAST+1] = t1[T_TOTAL]  - t1[0];

       mupc_reduce_r8(t1, tsum,  T_LAST+2, SMP_OPR_SUM, 0);
       mupc_reduce_r8(t1, tming, T_LAST+2, SMP_OPR_MIN, 0);
       mupc_reduce_r8(t1, tmaxg, T_LAST+2, SMP_OPR_MAX, 0);

       if (node == 0) {
          strcpy(t_recs[1], "total");
          strcpy(t_recs[2], "rhs");
          strcpy(t_recs[3], "xsolve");
          strcpy(t_recs[4], "ysolve");
          strcpy(t_recs[5], "zsolve");
          strcpy(t_recs[6], "bpack");
          strcpy(t_recs[7], "exch");
          strcpy(t_recs[8], "xcomm");
          strcpy(t_recs[9], "ycomm");
          strcpy(t_recs[10], "zcomm");
          strcpy(t_recs[11], " totcomp");
          strcpy(t_recs[0] , " totcomm");
          printf(" nprocs =%6d           minimum     maximum     %s\n", 
                 total_nodes, "average");
          for (j = 1; j <= T_LAST+2; j++) {
             i = j;
             if (i == T_LAST+2) i = 0;
             tsum[i] = tsum[i] / total_nodes;
             printf(" timer %2d(%-8s) :  %10.4f  %10.4f  %10.4f\n",
                    j, t_recs[i], tming[i], tmaxg[i], tsum[i]);
          }
       }
#endif
 done:
       upc_global_exit(0);

       return 0;

}
