
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include "header.h"
#include "upcnpb.h"

void z_solve() {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// this function performs the solution of the approximate factorization
// step in the z-direction for all five matrix components
// simultaneously. The Thomas algorithm is employed to solve the
// systems for the z-lines. Boundary conditions are non-periodic
//---------------------------------------------------------------------

       int i, j, k, stage, n, isize, jsize, kend, k1, k2,
               buffer_size, c, m, p, kstart, nid, ptr0;
       double  r1, r2, d, e, s[5], sm1, sm2,
                         fac1, fac2;
#define s(i) s[i-1]

#ifndef USE_BARRIER
       ptr0   = buffoff[2];
#endif

       TIMER_START(T_ZSOLVE);
//---------------------------------------------------------------------
// now do a sweep on a layer-by-layer basis, i.e. sweeping through cells
// on this node in the direction of increasing i for the forward sweep,
// and after that reversing the direction for the backsubstitution  
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//                          FORWARD ELIMINATION  
//---------------------------------------------------------------------
       for (stage = 1; stage <= ncells; stage++) {
          c         = slice(3,stage);

          kstart = 0;
          kend   = cell_size(3,c)-1;

          isize  = cell_size(1,c);
          jsize  = cell_size(2,c);

          buffer_size = (isize-start(1,c)-end(1,c)) * 
                        (jsize-start(2,c)-end(2,c));

          if (stage != 1) {


//---------------------------------------------------------------------
//            if this is not the first processor in this row of cells, 
//            receive data from predecessor containing the right hand
//            sides and the upper diagonal elements of the previous two rows
//---------------------------------------------------------------------

#ifdef USE_BARRIER
             TIMER_START(T_ZCOMM);
             upc_notify;
             TIMER_STOP(T_ZCOMM);
#endif


//---------------------------------------------------------------------
//            communication has already been started. 
//            compute the left hand side while waiting for the msg
//---------------------------------------------------------------------
             lhsz(c);

//---------------------------------------------------------------------
//            wait for pending communication to complete
//---------------------------------------------------------------------
             TIMER_START(T_ZCOMM);
             nid = predecessor(3);
#ifdef USE_BARRIER
      	     ptr0   = buffoff[buff_id];
      	     buff_id = 1 - buff_id;
             upc_wait;
#else
             sync_testwait(nid);
#endif
             upc_memget(&ul_buffer(1), &u_buffer(nid,ptr0),
                        sizeof(double)*22*buffer_size);
#ifndef USE_BARRIER
             sync_ackwait(nid);
#endif
             TIMER_STOP(T_ZCOMM);
         
//---------------------------------------------------------------------
//            unpack the buffer                                 
//---------------------------------------------------------------------
             k  = kstart;
             k1 = kstart + 1;
             n = 0;

//---------------------------------------------------------------------
//            create a running pointer
//---------------------------------------------------------------------
             p = 0;
             for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                   lhs(i,j,k,n+2,c) = lhs(i,j,k,n+2,c) -
                             ul_buffer(p+1) * lhs(i,j,k,n+1,c);
                   lhs(i,j,k,n+3,c) = lhs(i,j,k,n+3,c) -
                             ul_buffer(p+2) * lhs(i,j,k,n+1,c);
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) -
                             ul_buffer(p+2+m) * lhs(i,j,k,n+1,c);
                   }
                   d            = ul_buffer(p+6);
                   e            = ul_buffer(p+7);
                   for (m = 1; m <= 3; m++) {
                      s(m) = ul_buffer(p+7+m);
                   }
                   r1 = lhs(i,j,k,n+2,c);
                   lhs(i,j,k,n+3,c) = lhs(i,j,k,n+3,c) - d * r1;
                   lhs(i,j,k,n+4,c) = lhs(i,j,k,n+4,c) - e * r1;
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) - s(m) * r1;
                   }
                   r2 = lhs(i,j,k1,n+1,c);
                   lhs(i,j,k1,n+2,c) = lhs(i,j,k1,n+2,c) - d * r2;
                   lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) - e * r2;
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) - s(m) * r2;
                   }
                   p = p + 10;
                }
             }

             for (m = 4; m <= 5; m++) {
                n = (m-3)*5;
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      lhs(i,j,k,n+2,c) = lhs(i,j,k,n+2,c) -
                                ul_buffer(p+1) * lhs(i,j,k,n+1,c);
                      lhs(i,j,k,n+3,c) = lhs(i,j,k,n+3,c) -
                                ul_buffer(p+2) * lhs(i,j,k,n+1,c);
                      rhs(i,j,k,m,c)   = rhs(i,j,k,m,c) -
                                ul_buffer(p+3) * lhs(i,j,k,n+1,c);
                      d                = ul_buffer(p+4);
                      e                = ul_buffer(p+5);
                      s(m)             = ul_buffer(p+6);
                      r1 = lhs(i,j,k,n+2,c);
                      lhs(i,j,k,n+3,c) = lhs(i,j,k,n+3,c) - d * r1;
                      lhs(i,j,k,n+4,c) = lhs(i,j,k,n+4,c) - e * r1;
                      rhs(i,j,k,m,c)   = rhs(i,j,k,m,c) - s(m) * r1;
                      r2 = lhs(i,j,k1,n+1,c);
                      lhs(i,j,k1,n+2,c) = lhs(i,j,k1,n+2,c) - d * r2;
                      lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) - e * r2;
                      rhs(i,j,k1,m,c)   = rhs(i,j,k1,m,c) - s(m) * r2;
                      p = p + 6;
                   }
                }
             }

          } else {

//---------------------------------------------------------------------
//            if this IS the first cell, we still compute the lhs
//---------------------------------------------------------------------
             lhsz(c);
          }

//---------------------------------------------------------------------
//         perform the Thomas algorithm; first, FORWARD ELIMINATION     
//---------------------------------------------------------------------
          n = 0;

          for (k = kstart; k <= kend-2; k++) {
             for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                   k1 = k  + 1;
                   k2 = k  + 2;
                   fac1               = 1.e0/lhs(i,j,k,n+3,c);
                   lhs(i,j,k,n+4,c)   = fac1*lhs(i,j,k,n+4,c);
                   lhs(i,j,k,n+5,c)   = fac1*lhs(i,j,k,n+5,c);
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k,m,c) = fac1*rhs(i,j,k,m,c);
                   }
                   lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) -
                               lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+4,c);
                   lhs(i,j,k1,n+4,c) = lhs(i,j,k1,n+4,c) -
                               lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+5,c);
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) -
                               lhs(i,j,k1,n+2,c)*rhs(i,j,k,m,c);
                   }
                   lhs(i,j,k2,n+2,c) = lhs(i,j,k2,n+2,c) -
                               lhs(i,j,k2,n+1,c)*lhs(i,j,k,n+4,c);
                   lhs(i,j,k2,n+3,c) = lhs(i,j,k2,n+3,c) -
                               lhs(i,j,k2,n+1,c)*lhs(i,j,k,n+5,c);
                   for (m = 1; m <= 3; m++) {
                      rhs(i,j,k2,m,c) = rhs(i,j,k2,m,c) -
                               lhs(i,j,k2,n+1,c)*rhs(i,j,k,m,c);
                   }
                }
             }
          }

//---------------------------------------------------------------------
//         The last two rows in this grid block are a bit different, 
//         since they do not have two more rows available for the
//         elimination of off-diagonal entries
//---------------------------------------------------------------------
          k  = kend - 1;
          k1 = kend;
          for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
             for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                fac1               = 1.e0/lhs(i,j,k,n+3,c);
                lhs(i,j,k,n+4,c)   = fac1*lhs(i,j,k,n+4,c);
                lhs(i,j,k,n+5,c)   = fac1*lhs(i,j,k,n+5,c);
                for (m = 1; m <= 3; m++) {
                   rhs(i,j,k,m,c) = fac1*rhs(i,j,k,m,c);
                }
                lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) -
                            lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+4,c);
                lhs(i,j,k1,n+4,c) = lhs(i,j,k1,n+4,c) -
                            lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+5,c);
                for (m = 1; m <= 3; m++) {
                   rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) -
                            lhs(i,j,k1,n+2,c)*rhs(i,j,k,m,c);
                }
//---------------------------------------------------------------------
//               scale the last row immediately (some of this is
//               overkill in case this is the last cell)
//---------------------------------------------------------------------
                fac2               = 1.e0/lhs(i,j,k1,n+3,c);
                lhs(i,j,k1,n+4,c) = fac2*lhs(i,j,k1,n+4,c);
                lhs(i,j,k1,n+5,c) = fac2*lhs(i,j,k1,n+5,c)  ;
                for (m = 1; m <= 3; m++) {
                   rhs(i,j,k1,m,c) = fac2*rhs(i,j,k1,m,c);
                }
             }
          }

//---------------------------------------------------------------------
//         do the u+c and the u-c factors               
//---------------------------------------------------------------------
          for (m = 4; m <= 5; m++) {
             n = (m-3)*5;
             for (k = kstart; k <= kend-2; k++) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                   k1 = k  + 1;
                   k2 = k  + 2;
                   fac1               = 1.e0/lhs(i,j,k,n+3,c);
                   lhs(i,j,k,n+4,c)   = fac1*lhs(i,j,k,n+4,c);
                   lhs(i,j,k,n+5,c)   = fac1*lhs(i,j,k,n+5,c);
                   rhs(i,j,k,m,c) = fac1*rhs(i,j,k,m,c);
                   lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) -
                               lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+4,c);
                   lhs(i,j,k1,n+4,c) = lhs(i,j,k1,n+4,c) -
                               lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+5,c);
                   rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) -
                               lhs(i,j,k1,n+2,c)*rhs(i,j,k,m,c);
                   lhs(i,j,k2,n+2,c) = lhs(i,j,k2,n+2,c) -
                               lhs(i,j,k2,n+1,c)*lhs(i,j,k,n+4,c);
                   lhs(i,j,k2,n+3,c) = lhs(i,j,k2,n+3,c) -
                               lhs(i,j,k2,n+1,c)*lhs(i,j,k,n+5,c);
                   rhs(i,j,k2,m,c) = rhs(i,j,k2,m,c) -
                               lhs(i,j,k2,n+1,c)*rhs(i,j,k,m,c);
                }
             }
          }

//---------------------------------------------------------------------
//            And again the last two rows separately
//---------------------------------------------------------------------
             k  = kend - 1;
             k1 = kend;
             for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                fac1               = 1.e0/lhs(i,j,k,n+3,c);
                lhs(i,j,k,n+4,c)   = fac1*lhs(i,j,k,n+4,c);
                lhs(i,j,k,n+5,c)   = fac1*lhs(i,j,k,n+5,c);
                rhs(i,j,k,m,c)     = fac1*rhs(i,j,k,m,c);
                lhs(i,j,k1,n+3,c) = lhs(i,j,k1,n+3,c) -
                            lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+4,c);
                lhs(i,j,k1,n+4,c) = lhs(i,j,k1,n+4,c) -
                            lhs(i,j,k1,n+2,c)*lhs(i,j,k,n+5,c);
                rhs(i,j,k1,m,c)   = rhs(i,j,k1,m,c) -
                            lhs(i,j,k1,n+2,c)*rhs(i,j,k,m,c);
//---------------------------------------------------------------------
//               Scale the last row immediately (some of this is overkill
//               if this is the last cell)
//---------------------------------------------------------------------
                fac2               = 1.e0/lhs(i,j,k1,n+3,c);
                lhs(i,j,k1,n+4,c) = fac2*lhs(i,j,k1,n+4,c);
                lhs(i,j,k1,n+5,c) = fac2*lhs(i,j,k1,n+5,c);
                rhs(i,j,k1,m,c)   = fac2*rhs(i,j,k1,m,c);

             }
          }
       }

//---------------------------------------------------------------------
//         send information to the next processor, except when this
//         is the last grid block,
//---------------------------------------------------------------------

          if (stage != ncells) {

//---------------------------------------------------------------------
//            create a running pointer for the send buffer  
//---------------------------------------------------------------------
             p = 0;
             n = 0;
             for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                   for (k = kend-1; k <= kend; k++) {
                      ul_buffer(p+1) = lhs(i,j,k,n+4,c);
                      ul_buffer(p+2) = lhs(i,j,k,n+5,c);
                      for (m = 1; m <= 3; m++) {
                         ul_buffer(p+2+m) = rhs(i,j,k,m,c);
                      }
                      p = p+5;
                   }
                }
             }

             for (m = 4; m <= 5; m++) {
                n = (m-3)*5;
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      for (k = kend-1; k <= kend; k++) {
                         ul_buffer(p+1) = lhs(i,j,k,n+4,c);
                         ul_buffer(p+2) = lhs(i,j,k,n+5,c);
                         ul_buffer(p+3) = rhs(i,j,k,m,c);
                         p = p + 3;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
// send data to next phase
//---------------------------------------------------------------------
             TIMER_START(T_ZCOMM);
#ifdef USE_BARRIER
      	     ptr0   = buffoff[buff_id];
#else
             nid = successor(3);
             sync_testnotify(nid);
#endif
             upc_memput(&u_buffer(node,ptr0), &ul_buffer(1),
                        sizeof(double)*22*buffer_size);
#ifndef USE_BARRIER
             sync_setnotify(nid);
#endif
             TIMER_STOP(T_ZCOMM);

          }
       }

//---------------------------------------------------------------------
//      now go in the reverse direction                      
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//                         BACKSUBSTITUTION 
//---------------------------------------------------------------------
       for (stage = ncells; stage >= 1; stage--) {
          c = slice(3,stage);

          kstart = 0;
          kend   = cell_size(3,c)-1;

          isize  = cell_size(1,c);
          jsize  = cell_size(2,c);

          buffer_size = (isize-start(1,c)-end(1,c)) * 
                        (jsize-start(2,c)-end(2,c));

          if (stage != ncells) {

//---------------------------------------------------------------------
//            if this is not the starting cell in this row of cells, 
//            wait for a message to be received, containing the 
//            solution of the previous two stations     
//---------------------------------------------------------------------

#ifdef USE_BARRIER
             TIMER_START(T_ZCOMM);
             upc_notify;
             TIMER_STOP(T_ZCOMM);
#endif


//---------------------------------------------------------------------
//            communication has already been started
//            while waiting, do the  block-diagonal inversion for the 
//            cell that was just finished                
//---------------------------------------------------------------------

             tzetar(slice(3,stage+1));

//---------------------------------------------------------------------
//            wait for pending communication to complete
//---------------------------------------------------------------------
             TIMER_START(T_ZCOMM);
             nid = successor(3);
#ifdef USE_BARRIER
      	     ptr0   = buffoff[buff_id];
      	     buff_id = 1 - buff_id;
             upc_wait;
#else
             sync_testwait(nid);
#endif
             upc_memget(&ul_buffer(1), &u_buffer(nid,ptr0),
                        sizeof(double)*10*buffer_size);
#ifndef USE_BARRIER
             sync_ackwait(nid);
#endif
             TIMER_STOP(T_ZCOMM);

//---------------------------------------------------------------------
//            unpack the buffer for the first three factors         
//---------------------------------------------------------------------
             n = 0;
             p = 0;
             k  = kend;
             k1 = k - 1;
             for (m = 1; m <= 3; m++) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      sm1 = ul_buffer(p+1);
                      sm2 = ul_buffer(p+2);
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) - 
                              lhs(i,j,k,n+4,c)*sm1 -
                              lhs(i,j,k,n+5,c)*sm2;
                      rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) -
                              lhs(i,j,k1,n+4,c) * rhs(i,j,k,m,c) - 
                              lhs(i,j,k1,n+5,c) * sm1;
                      p = p + 2;
                   }
                }
             }

//---------------------------------------------------------------------
//            now unpack the buffer for the remaining two factors
//---------------------------------------------------------------------
             for (m = 4; m <= 5; m++) {
                n = (m-3)*5;
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      sm1 = ul_buffer(p+1);
                      sm2 = ul_buffer(p+2);
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) - 
                              lhs(i,j,k,n+4,c)*sm1 -
                              lhs(i,j,k,n+5,c)*sm2;
                      rhs(i,j,k1,m,c) = rhs(i,j,k1,m,c) -
                              lhs(i,j,k1,n+4,c) * rhs(i,j,k,m,c) - 
                              lhs(i,j,k1,n+5,c) * sm1;
                      p = p + 2;
                   }
                }
             }

          } else {

//---------------------------------------------------------------------
//            now we know this is the first grid block on the back sweep,
//            so we don't need a message to start the substitution. 
//---------------------------------------------------------------------

             k  = kend - 1;
             k1 = kend;
             n = 0;
             for (m = 1; m <= 3; m++) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) -
                                   lhs(i,j,k,n+4,c)*rhs(i,j,k1,m,c);
                   }
                }
             }

             for (m = 4; m <= 5; m++) {
                n = (m-3)*5;
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) -
                                   lhs(i,j,k,n+4,c)*rhs(i,j,k1,m,c);
                   }
                }
             }
          }

//---------------------------------------------------------------------
//         Whether or not this is the last processor, we always have
//         to complete the back-substitution 
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//         The first three factors
//---------------------------------------------------------------------
          n = 0;
          for (m = 1; m <= 3; m++) {
             for (k = kend-2; k >= kstart; k--) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      k1 = k  + 1;
                      k2 = k  + 2;
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) - 
                                lhs(i,j,k,n+4,c)*rhs(i,j,k1,m,c) -
                                lhs(i,j,k,n+5,c)*rhs(i,j,k2,m,c);
                   }
                }
             }
          }

//---------------------------------------------------------------------
//         And the remaining two
//---------------------------------------------------------------------
          for (m = 4; m <= 5; m++) {
             n = (m-3)*5;
             for (k = kend-2; k >= kstart; k--) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      k1 = k  + 1;
                      k2 = k  + 2;
                      rhs(i,j,k,m,c) = rhs(i,j,k,m,c) - 
                                lhs(i,j,k,n+4,c)*rhs(i,j,k1,m,c) -
                                lhs(i,j,k,n+5,c)*rhs(i,j,k2,m,c);
                   }
                }
             }
          }

//---------------------------------------------------------------------
//         send on information to the previous processor, if needed
//---------------------------------------------------------------------
          if (stage !=  1) {
             k  = kstart;
             k1 = kstart + 1;
             p = 0;
             for (m = 1; m <= 5; m++) {
                for (j = start(2,c); j <= jsize-end(2,c)-1; j++) {
                   for (i = start(1,c); i <= isize-end(1,c)-1; i++) {
                      ul_buffer(p+1) = rhs(i,j,k,m,c);
                      ul_buffer(p+2) = rhs(i,j,k1,m,c);
                      p = p + 2;
                   }
                }
             }

//---------------------------------------------------------------------
//            pack and send the buffer
//---------------------------------------------------------------------
             TIMER_START(T_ZCOMM);
#ifdef USE_BARRIER
      	     ptr0   = buffoff[buff_id];
#else
             nid = predecessor(3);
             sync_testnotify(nid);
#endif
             upc_memput(&u_buffer(node,ptr0), &ul_buffer(1),
                        sizeof(double)*10*buffer_size);
#ifndef USE_BARRIER
             sync_setnotify(nid);
#endif
             TIMER_STOP(T_ZCOMM);

          }

//---------------------------------------------------------------------
//         If this was the last stage, do the block-diagonal inversion
//---------------------------------------------------------------------
          if (stage == 1) tzetar(c);

       }

       TIMER_STOP(T_ZSOLVE);

       return;
#undef s
}
    






