
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <stdio.h>
#include <unistd.h>
#include "header.h"
#include "upcnpb.h"
#include <upc_io.h>

static upc_file_t *btio_fp;
static void acc_sub_norms(int idump_cur);

upc_off_t iseek;

struct upc_local_memvec memvec[MAXCELLS*JMAX*KMAX];
struct upc_filevec filevec[MAXCELLS*JMAX*KMAX];
size_t nvecs, iossize;


void setup_btio(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int m;

      int count, jio, kio, cio;
      upc_off_t offset;

      if (node == root) {
         /* remove old file if exists */
         unlink(filenm);
      }
      upc_barrier;

      btio_fp = upc_all_fopen(filenm, 
                              UPC_INDIVIDUAL_FP|UPC_RDWR|UPC_CREATE, 
                              0, NULL);

      if (!btio_fp) {
          printf(" Error opening output file\n");
          upc_global_exit(1);
      }

      nvecs = 0;
      for (cio = 1; cio <= ncells; cio++) {
          for (kio = 0; kio <= cell_size(3,cio)-1; kio++) {
              for (jio = 0; jio <= cell_size(2,cio)-1; jio++) {
                  iseek=5*(cell_low(1,cio) +
                         PROBLEM_SIZE*((cell_low(2,cio)+jio) +
                          PROBLEM_SIZE*(cell_low(3,cio)+kio)));

                  count=5*cell_size(1,cio);

                  filevec[nvecs].offset = iseek*sizeof(double);
                  filevec[nvecs].len = count*sizeof(double);
                  memvec[nvecs].baseaddr = &u(1,0,jio,kio,cio);
                  memvec[nvecs].len = count*sizeof(double);
                  nvecs++;
              }
          }
      }

      iossize = sizeof(double)*5*PROBLEM_SIZE*PROBLEM_SIZE*PROBLEM_SIZE;

      for (m = 1; m <= 5; m++) {
         xce_sub(m) = 0.e0;
      }

      idump_sub = 0;

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void output_timestep(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int count;
      upc_off_t offset;

      offset = upc_all_fwrite_list_local(btio_fp,
                  nvecs, memvec, nvecs, filevec, 0);
      if (offset == -1) {
          printf(" Error writing to file\n");
          upc_global_exit(-1);
      }

      idump_sub = idump_sub + 1;
      for (count = 0; count < nvecs; count++)
         filevec[count].offset += iossize;

      if (rd_interval > 0) {
         if (idump_sub >= rd_interval) {

            acc_sub_norms(idump+1);

            for (count = 0; count < nvecs; count++)
               filevec[count].offset -= iossize*idump_sub;
            idump_sub = 0;
         }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

static void acc_sub_norms(int idump_cur) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int count, ii, m, ichunk;
      upc_off_t offset;

      double xce_single[5];

      for (count = 0; count < nvecs; count++)
         filevec[count].offset -= iossize*idump_sub;

      ichunk = idump_cur - idump_sub + 1;
      for (ii = 0; ii <= idump_sub-1; ii++) {
        offset = upc_all_fread_list_local(btio_fp,
                    nvecs, memvec, nvecs, filevec, 0);
        if (offset == -1) {
           printf(" Error reading back file\n");
           upc_global_exit(-1);
        }

        for (count = 0; count < nvecs; count++)
           filevec[count].offset += iossize;

        if (node == root) printf(" Reading data set %4d\n", ii+ichunk);

        error_norm(xce_single);
        for (m = 1; m <= 5; m++) {
           xce_sub(m) = xce_sub(m) + xce_single[m-1];
        }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void btio_cleanup(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      if (upc_all_fclose(btio_fp) == -1) {
         printf("Error closing file\n");
         upc_global_exit(-1);
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void accumulate_norms(double xce_acc[]) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int m;

      if (rd_interval <= 0) {

      btio_fp = upc_all_fopen(filenm, UPC_INDIVIDUAL_FP|UPC_RDONLY, 0, NULL);
      if (!btio_fp) {
         printf("Error opening file for read\n");
         upc_global_exit(1);
      }

//     clear the last time step

      clear_timestep();

//     read back the time steps and accumulate norms

      acc_sub_norms(idump);

      upc_all_fclose(btio_fp);
      }

      for (m = 1; m <= 5; m++) {
         xce_acc[m-1] = xce_sub(m) / (double)(idump);
      }

      return;
}

