//---------------------------------------------------------------------
//---------------------------------------------------------------------
#define WORK_LHS_VEC
#include "header.h"
#include "upcnpb.h"


extern void y_unpack_solve_info(int c);
extern void y_send_solve_info(int c);
extern void y_send_backsub_info(int c);
extern void y_unpack_backsub_info(int c);
extern void y_backsubstitute(int first, int last, int c);
extern void y_solve_cell(int first, int last, int c);

void y_solve() {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     Performs line solves in Y direction by first factoring
//     the block-tridiagonal matrix into an upper triangular matrix, 
//     and then performing back substitution to solve for the unknown
//     vectors of each line.  
//     
//     Make sure we treat elements zero to cell_size in the direction
//     of the sweep.
//---------------------------------------------------------------------

      int  c, stage,
           first, last;

//---------------------------------------------------------------------
//     in our terminology stage is the number of the cell in the y-direction
//     i.e. stage = 1 means the start of the line stage=ncells means end
//---------------------------------------------------------------------
      for (stage = 1; stage <= ncells; stage++) {
         c = slice(2,stage);

//---------------------------------------------------------------------
//     set last-cell flag
//---------------------------------------------------------------------
         if (stage == ncells) {
            last = 1;
         } else {
            last = 0;
         }

         if (stage == 1) {
//---------------------------------------------------------------------
//     This is the first cell, so solve without receiving data
//---------------------------------------------------------------------
            first = 1;
//            call lhsy(c)
            y_solve_cell(first,last,c);
         } else {
//---------------------------------------------------------------------
//     Not the first cell of this line, so receive info from
//     processor working on preceeding cell
//---------------------------------------------------------------------
            first = 0;
//            y_receive_solve_info(&recv_id,c);
//---------------------------------------------------------------------
//     overlap computations and communications
//---------------------------------------------------------------------
//            call lhsy(c)
//---------------------------------------------------------------------
//     wait for completion
//---------------------------------------------------------------------
#ifdef USE_BARRIER
            upc_barrier;
#endif
//---------------------------------------------------------------------
//     install C'(jstart+1) and rhs'(jstart+1) to be used in this cell
//---------------------------------------------------------------------
            y_unpack_solve_info(c);
            y_solve_cell(first,last,c);
         }

         if (last == 0) y_send_solve_info(c);
      }

//---------------------------------------------------------------------
//     now perform backsubstitution in reverse direction
//---------------------------------------------------------------------
      for (stage = ncells; stage >= 1; stage--) {
         c = slice(2,stage);
         first = 0;
         last = 0;
         if (stage == 1) first = 1;
         if (stage == ncells) {
            last = 1;
//---------------------------------------------------------------------
//     last cell, so perform back substitute without waiting
//---------------------------------------------------------------------
            y_backsubstitute(first, last,c);
         } else {
//            y_receive_backsub_info(&recv_id,c);
#ifdef USE_BARRIER
            upc_barrier;
#endif
            y_unpack_backsub_info(c);
            y_backsubstitute(first,last,c);
         }
         if (first == 0) y_send_backsub_info(c);
      }


      return;
}
      
//---------------------------------------------------------------------
//---------------------------------------------------------------------
      
void y_unpack_solve_info(int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     unpack C'(-1) and rhs'(-1) for
//     all i and k
//---------------------------------------------------------------------

      int i,k,m,n,ptr,jstart,bsize,nid ;

      nid = predecessor(2);
      bsize = (BLOCK_SIZE+1)*BLOCK_SIZE*IMAX*KMAX;
#ifndef USE_BARRIER
      ptr = buffoff[1];
      sync_testwait(nid);
#else
      ptr = buffoff[buff_id];
      buff_id = 1 - buff_id;
#endif
      upc_memget(&ul_buffer(1), &u_buffer(nid,ptr), sizeof(double)*bsize);
#ifndef USE_BARRIER
      sync_ackwait(nid);
#endif

      jstart = 0;
      ptr = 0;
      for (k = 0; k < KMAX; k++) {
         for (i = 0; i < IMAX; i++) {
            for (m = 1; m <= BLOCK_SIZE; m++) {
               for (n = 1; n <= BLOCK_SIZE; n++) {
                  lhsc(m,n,i,jstart-1,k,c) = ul_buffer(ptr+n);
               }
               ptr = ptr+BLOCK_SIZE;
            }
            for (n = 1; n <= BLOCK_SIZE; n++) {
               rhs(n,i,jstart-1,k,c) = ul_buffer(ptr+n);
            }
            ptr = ptr+BLOCK_SIZE;
         }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------
      
void y_send_solve_info(int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     pack up and send C'(jend) and rhs'(jend) for
//     all i and k
//---------------------------------------------------------------------

      int i,k,m,n,jsize,ptr,bsize,nid;

      jsize = cell_size(2,c)-1;

//---------------------------------------------------------------------
//     pack up buffer
//---------------------------------------------------------------------
      ptr = 0;
      for (k = 0; k < KMAX; k++) {
         for (i = 0; i < IMAX; i++) {
            for (m = 1; m <= BLOCK_SIZE; m++) {
               for (n = 1; n <= BLOCK_SIZE; n++) {
                  ul_buffer(ptr+n) = lhsc(m,n,i,jsize,k,c);
               }
               ptr = ptr+BLOCK_SIZE;
            }
            for (n = 1; n <= BLOCK_SIZE; n++) {
               ul_buffer(ptr+n) = rhs(n,i,jsize,k,c);
            }
            ptr = ptr+BLOCK_SIZE;
         }
      }

//---------------------------------------------------------------------
//     send buffer 
//---------------------------------------------------------------------
      bsize = (BLOCK_SIZE+1)*BLOCK_SIZE*IMAX*KMAX;
#ifndef USE_BARRIER
      ptr = buffoff[1];
      nid = successor(2);
      sync_testnotify(nid);
#else
      ptr = buffoff[buff_id];
#endif
      upc_memput(&u_buffer(node,ptr), &ul_buffer(1), sizeof(double)*bsize);
#ifndef USE_BARRIER
      sync_setnotify(nid);
#endif

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void y_send_backsub_info(int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     pack up and send U(jstart) for all i and k
//---------------------------------------------------------------------

      int i,k,n,ptr,jstart,bsize,nid;

//---------------------------------------------------------------------
//     Send element 0 to previous processor
//---------------------------------------------------------------------
      jstart = 0;
      ptr = 0;
      for (k = 0; k < KMAX; k++) {
         for (i = 0; i < IMAX; i++) {
            for (n = 1; n <= BLOCK_SIZE; n++) {
               ul_buffer(ptr+n) = rhs(n,i,jstart,k,c);
            }
            ptr = ptr+BLOCK_SIZE;
         }
      }

      bsize = BLOCK_SIZE*IMAX*KMAX;
#ifndef USE_BARRIER
      ptr = buffoff[1];
      nid = predecessor(2);
      sync_testnotify(nid);
#else
      ptr = buffoff[buff_id];
#endif
      upc_memput(&u_buffer(node,ptr), &ul_buffer(1), sizeof(double)*bsize);
#ifndef USE_BARRIER
      sync_setnotify(nid);
#endif

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void y_unpack_backsub_info(int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     unpack U(jsize) for all i and k
//---------------------------------------------------------------------

      int i,k,n,ptr,bsize,nid ;

      nid = successor(2);
      bsize = BLOCK_SIZE*IMAX*KMAX;
#ifndef USE_BARRIER
      ptr = buffoff[1];
      sync_testwait(nid);
#else
      ptr = buffoff[buff_id];
      buff_id = 1 - buff_id;
#endif
      upc_memget(&ul_buffer(1), &u_buffer(nid,ptr), sizeof(double)*bsize);
#ifndef USE_BARRIER
      sync_ackwait(nid);
#endif

      ptr = 0;
      for (k = 0; k < KMAX; k++) {
         for (i = 0; i < IMAX; i++) {
            for (n = 1; n <= BLOCK_SIZE; n++) {
               backsub_info(n,i,k,c) = ul_buffer(ptr+n);
            }
            ptr = ptr+BLOCK_SIZE;
         }
      }

      return;
}
      
//---------------------------------------------------------------------
//---------------------------------------------------------------------

void y_backsubstitute(int first, int last, int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     back solve: if last cell, then generate U(jsize)=rhs(jsize)
//     else assume U(jsize) is loaded in un pack backsub_info
//     so just use it
//     after call u(jstart) will be sent to next cell
//---------------------------------------------------------------------

      int i, k;
      int m,n,j,jsize,isize,ksize,jstart;
      
      jstart = 0;
      isize = cell_size(1,c)-end(1,c)-1      ;
      jsize = cell_size(2,c)-1;
      ksize = cell_size(3,c)-end(3,c)-1;
      if (last == 0) {
         for (k = start(3,c); k <= ksize; k++) {
            for (i = start(1,c); i <= isize; i++) {
//---------------------------------------------------------------------
//     U(jsize) uses info from previous cell if not last cell
//---------------------------------------------------------------------
               for (m = 1; m <= BLOCK_SIZE; m++) {
                  for (n = 1; n <= BLOCK_SIZE; n++) {
                     rhs(m,i,jsize,k,c) = rhs(m,i,jsize,k,c) 
                          - lhsc(m,n,i,jsize,k,c)*
                          backsub_info(n,i,k,c);
                  }
               }
            }
         }
      }
      for (k = start(3,c); k <= ksize; k++) {
         for (j = jsize-1; j >= jstart; j--) {
            for (i = start(1,c); i <= isize; i++) {
               for (m = 1; m <= BLOCK_SIZE; m++) {
                  for (n = 1; n <= BLOCK_SIZE; n++) {
                     rhs(m,i,j,k,c) = rhs(m,i,j,k,c) 
                          - lhsc(m,n,i,j,k,c)*rhs(n,i,j+1,k,c);
                  }
               }
            }
         }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void y_solve_cell(int first,int last,int c) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     performs guaussian elimination on this cell.
//     
//     assumes that unpacking routines for non-first cells 
//     preload C' and rhs' from previous cell.
//     
//     assumed send happens outside this routine, but that
//     c'(JMAX) and rhs'(JMAX) will be sent to next cell
//---------------------------------------------------------------------

      int i,j,k,m,n,isize,ksize,jsize,jstart;

      jstart = 0;
      isize = cell_size(1,c)-end(1,c)-1;
      jsize = cell_size(2,c)-1;
      ksize = cell_size(3,c)-end(3,c)-1;

//---------------------------------------------------------------------
//     zero the left hand side for starters
//     set diagonal values to 1. This is overkill, but convenient
//---------------------------------------------------------------------
      for (i = 0; i <= isize; i++) {
         for (m = 1; m <= 5; m++) {
            for (n = 1; n <= 5; n++) {
               lhsa(m,n,i,0) = 0.0e0;
               lhsb(m,n,i,0) = 0.0e0;
               lhsa(m,n,i,jsize) = 0.0e0;
               lhsb(m,n,i,jsize) = 0.0e0;
            }
            lhsb(m,m,i,0) = 1.0e0;
            lhsb(m,m,i,jsize) = 1.0e0;
         }
      }

      for (k = start(3,c); k <= ksize; k++) {

//---------------------------------------------------------------------
//     This function computes the left hand side for the three y-factors 
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//     Compute the indices for storing the tri-diagonal matrix;
//     determine a (labeled f) and n jacobians for cell c
//---------------------------------------------------------------------

         for (j = start(2,c)-1; j <= cell_size(2,c)-end(2,c); j++) {
            for (i = start(1,c); i <= isize; i++) {

               tmp1 = 1.0e0 / u(1,i,j,k,c);
               tmp2 = tmp1 * tmp1;
               tmp3 = tmp1 * tmp2;

               fjac(1,1,i,j) = 0.0e+00;
               fjac(1,2,i,j) = 0.0e+00;
               fjac(1,3,i,j) = 1.0e+00;
               fjac(1,4,i,j) = 0.0e+00;
               fjac(1,5,i,j) = 0.0e+00;

               fjac(2,1,i,j) = - ( u(2,i,j,k,c)*u(3,i,j,k,c) )
                    * tmp2;
               fjac(2,2,i,j) = u(3,i,j,k,c) * tmp1;
               fjac(2,3,i,j) = u(2,i,j,k,c) * tmp1;
               fjac(2,4,i,j) = 0.0e+00;
               fjac(2,5,i,j) = 0.0e+00;

               fjac(3,1,i,j) = - ( u(3,i,j,k,c)*u(3,i,j,k,c)*tmp2)
                    + c2 * qs(i,j,k,c);
               fjac(3,2,i,j) = - c2 *  u(2,i,j,k,c) * tmp1;
               fjac(3,3,i,j) = ( 2.0e+00 - c2 )
                    *  u(3,i,j,k,c) * tmp1 ;
               fjac(3,4,i,j) = - c2 * u(4,i,j,k,c) * tmp1 ;
               fjac(3,5,i,j) = c2;

               fjac(4,1,i,j) = - ( u(3,i,j,k,c)*u(4,i,j,k,c) )
                    * tmp2;
               fjac(4,2,i,j) = 0.0e+00;
               fjac(4,3,i,j) = u(4,i,j,k,c) * tmp1;
               fjac(4,4,i,j) = u(3,i,j,k,c) * tmp1;
               fjac(4,5,i,j) = 0.0e+00;

               fjac(5,1,i,j) = ( c2 * 2.0e0 * qs(i,j,k,c)
                    - c1 * u(5,i,j,k,c) * tmp1 ) 
                    * u(3,i,j,k,c) * tmp1 ;
               fjac(5,2,i,j) = - c2 * u(2,i,j,k,c)*u(3,i,j,k,c) 
                    * tmp2;
               fjac(5,3,i,j) = c1 * u(5,i,j,k,c) * tmp1 
                    - c2 * ( qs(i,j,k,c)
                    + u(3,i,j,k,c)*u(3,i,j,k,c) * tmp2 );
               fjac(5,4,i,j) = - c2 * ( u(3,i,j,k,c)*u(4,i,j,k,c) )
                    * tmp2;
               fjac(5,5,i,j) = c1 * u(3,i,j,k,c) * tmp1 ;

               njac(1,1,i,j) = 0.0e+00;
               njac(1,2,i,j) = 0.0e+00;
               njac(1,3,i,j) = 0.0e+00;
               njac(1,4,i,j) = 0.0e+00;
               njac(1,5,i,j) = 0.0e+00;

               njac(2,1,i,j) = - c3c4 * tmp2 * u(2,i,j,k,c);
               njac(2,2,i,j) =   c3c4 * tmp1;
               njac(2,3,i,j) =   0.0e+00;
               njac(2,4,i,j) =   0.0e+00;
               njac(2,5,i,j) =   0.0e+00;

               njac(3,1,i,j) = - con43 * c3c4 * tmp2 * u(3,i,j,k,c);
               njac(3,2,i,j) =   0.0e+00;
               njac(3,3,i,j) =   con43 * c3c4 * tmp1;
               njac(3,4,i,j) =   0.0e+00;
               njac(3,5,i,j) =   0.0e+00;

               njac(4,1,i,j) = - c3c4 * tmp2 * u(4,i,j,k,c);
               njac(4,2,i,j) =   0.0e+00;
               njac(4,3,i,j) =   0.0e+00;
               njac(4,4,i,j) =   c3c4 * tmp1;
               njac(4,5,i,j) =   0.0e+00;

               njac(5,1,i,j) = - (  c3c4
                    - c1345 ) * tmp3 * SQR(u(2,i,j,k,c))
                    - ( con43 * c3c4
                    - c1345 ) * tmp3 * SQR(u(3,i,j,k,c))
                    - ( c3c4 - c1345 ) * tmp3 * SQR(u(4,i,j,k,c))
                    - c1345 * tmp2 * u(5,i,j,k,c);

               njac(5,2,i,j) = (  c3c4 - c1345 ) * tmp2 * u(2,i,j,k,c);
               njac(5,3,i,j) = ( con43 * c3c4
                    - c1345 ) * tmp2 * u(3,i,j,k,c);
               njac(5,4,i,j) = ( c3c4 - c1345 ) * tmp2 * u(4,i,j,k,c);
               njac(5,5,i,j) = ( c1345 ) * tmp1;

            }
         }

//---------------------------------------------------------------------
//     now joacobians set, so form left hand side in y direction
//---------------------------------------------------------------------
         for (j = start(2,c); j <= jsize-end(2,c); j++) {
            for (i = start(1,c); i <= isize; i++) {

               tmp1 = dt * ty1;
               tmp2 = dt * ty2;

               lhsa(1,1,i,j) = - tmp2 * fjac(1,1,i,j-1)
                    - tmp1 * njac(1,1,i,j-1)
                    - tmp1 * dy1 ;
               lhsa(1,2,i,j) = - tmp2 * fjac(1,2,i,j-1)
                    - tmp1 * njac(1,2,i,j-1);
               lhsa(1,3,i,j) = - tmp2 * fjac(1,3,i,j-1)
                    - tmp1 * njac(1,3,i,j-1);
               lhsa(1,4,i,j) = - tmp2 * fjac(1,4,i,j-1)
                    - tmp1 * njac(1,4,i,j-1);
               lhsa(1,5,i,j) = - tmp2 * fjac(1,5,i,j-1)
                    - tmp1 * njac(1,5,i,j-1);

               lhsa(2,1,i,j) = - tmp2 * fjac(2,1,i,j-1)
                    - tmp1 * njac(2,1,i,j-1);
               lhsa(2,2,i,j) = - tmp2 * fjac(2,2,i,j-1)
                    - tmp1 * njac(2,2,i,j-1)
                    - tmp1 * dy2;
               lhsa(2,3,i,j) = - tmp2 * fjac(2,3,i,j-1)
                    - tmp1 * njac(2,3,i,j-1);
               lhsa(2,4,i,j) = - tmp2 * fjac(2,4,i,j-1)
                    - tmp1 * njac(2,4,i,j-1);
               lhsa(2,5,i,j) = - tmp2 * fjac(2,5,i,j-1)
                    - tmp1 * njac(2,5,i,j-1);

               lhsa(3,1,i,j) = - tmp2 * fjac(3,1,i,j-1)
                    - tmp1 * njac(3,1,i,j-1);
               lhsa(3,2,i,j) = - tmp2 * fjac(3,2,i,j-1)
                    - tmp1 * njac(3,2,i,j-1);
               lhsa(3,3,i,j) = - tmp2 * fjac(3,3,i,j-1)
                    - tmp1 * njac(3,3,i,j-1)
                    - tmp1 * dy3 ;
               lhsa(3,4,i,j) = - tmp2 * fjac(3,4,i,j-1)
                    - tmp1 * njac(3,4,i,j-1);
               lhsa(3,5,i,j) = - tmp2 * fjac(3,5,i,j-1)
                    - tmp1 * njac(3,5,i,j-1);

               lhsa(4,1,i,j) = - tmp2 * fjac(4,1,i,j-1)
                    - tmp1 * njac(4,1,i,j-1);
               lhsa(4,2,i,j) = - tmp2 * fjac(4,2,i,j-1)
                    - tmp1 * njac(4,2,i,j-1);
               lhsa(4,3,i,j) = - tmp2 * fjac(4,3,i,j-1)
                    - tmp1 * njac(4,3,i,j-1);
               lhsa(4,4,i,j) = - tmp2 * fjac(4,4,i,j-1)
                    - tmp1 * njac(4,4,i,j-1)
                    - tmp1 * dy4;
               lhsa(4,5,i,j) = - tmp2 * fjac(4,5,i,j-1)
                    - tmp1 * njac(4,5,i,j-1);

               lhsa(5,1,i,j) = - tmp2 * fjac(5,1,i,j-1)
                    - tmp1 * njac(5,1,i,j-1);
               lhsa(5,2,i,j) = - tmp2 * fjac(5,2,i,j-1)
                    - tmp1 * njac(5,2,i,j-1);
               lhsa(5,3,i,j) = - tmp2 * fjac(5,3,i,j-1)
                    - tmp1 * njac(5,3,i,j-1);
               lhsa(5,4,i,j) = - tmp2 * fjac(5,4,i,j-1)
                    - tmp1 * njac(5,4,i,j-1);
               lhsa(5,5,i,j) = - tmp2 * fjac(5,5,i,j-1)
                    - tmp1 * njac(5,5,i,j-1)
                    - tmp1 * dy5;

               lhsb(1,1,i,j) = 1.0e+00
                    + tmp1 * 2.0e+00 * njac(1,1,i,j)
                    + tmp1 * 2.0e+00 * dy1;
               lhsb(1,2,i,j) = tmp1 * 2.0e+00 * njac(1,2,i,j);
               lhsb(1,3,i,j) = tmp1 * 2.0e+00 * njac(1,3,i,j);
               lhsb(1,4,i,j) = tmp1 * 2.0e+00 * njac(1,4,i,j);
               lhsb(1,5,i,j) = tmp1 * 2.0e+00 * njac(1,5,i,j);

               lhsb(2,1,i,j) = tmp1 * 2.0e+00 * njac(2,1,i,j);
               lhsb(2,2,i,j) = 1.0e+00
                    + tmp1 * 2.0e+00 * njac(2,2,i,j)
                    + tmp1 * 2.0e+00 * dy2;
               lhsb(2,3,i,j) = tmp1 * 2.0e+00 * njac(2,3,i,j);
               lhsb(2,4,i,j) = tmp1 * 2.0e+00 * njac(2,4,i,j);
               lhsb(2,5,i,j) = tmp1 * 2.0e+00 * njac(2,5,i,j);

               lhsb(3,1,i,j) = tmp1 * 2.0e+00 * njac(3,1,i,j);
               lhsb(3,2,i,j) = tmp1 * 2.0e+00 * njac(3,2,i,j);
               lhsb(3,3,i,j) = 1.0e+00
                    + tmp1 * 2.0e+00 * njac(3,3,i,j)
                    + tmp1 * 2.0e+00 * dy3;
               lhsb(3,4,i,j) = tmp1 * 2.0e+00 * njac(3,4,i,j);
               lhsb(3,5,i,j) = tmp1 * 2.0e+00 * njac(3,5,i,j);

               lhsb(4,1,i,j) = tmp1 * 2.0e+00 * njac(4,1,i,j);
               lhsb(4,2,i,j) = tmp1 * 2.0e+00 * njac(4,2,i,j);
               lhsb(4,3,i,j) = tmp1 * 2.0e+00 * njac(4,3,i,j);
               lhsb(4,4,i,j) = 1.0e+00
                    + tmp1 * 2.0e+00 * njac(4,4,i,j)
                    + tmp1 * 2.0e+00 * dy4;
               lhsb(4,5,i,j) = tmp1 * 2.0e+00 * njac(4,5,i,j);

               lhsb(5,1,i,j) = tmp1 * 2.0e+00 * njac(5,1,i,j);
               lhsb(5,2,i,j) = tmp1 * 2.0e+00 * njac(5,2,i,j);
               lhsb(5,3,i,j) = tmp1 * 2.0e+00 * njac(5,3,i,j);
               lhsb(5,4,i,j) = tmp1 * 2.0e+00 * njac(5,4,i,j);
               lhsb(5,5,i,j) = 1.0e+00
                    + tmp1 * 2.0e+00 * njac(5,5,i,j) 
                    + tmp1 * 2.0e+00 * dy5;

               lhsc(1,1,i,j,k,c) =  tmp2 * fjac(1,1,i,j+1)
                    - tmp1 * njac(1,1,i,j+1)
                    - tmp1 * dy1;
               lhsc(1,2,i,j,k,c) =  tmp2 * fjac(1,2,i,j+1)
                    - tmp1 * njac(1,2,i,j+1);
               lhsc(1,3,i,j,k,c) =  tmp2 * fjac(1,3,i,j+1)
                    - tmp1 * njac(1,3,i,j+1);
               lhsc(1,4,i,j,k,c) =  tmp2 * fjac(1,4,i,j+1)
                    - tmp1 * njac(1,4,i,j+1);
               lhsc(1,5,i,j,k,c) =  tmp2 * fjac(1,5,i,j+1)
                    - tmp1 * njac(1,5,i,j+1);

               lhsc(2,1,i,j,k,c) =  tmp2 * fjac(2,1,i,j+1)
                    - tmp1 * njac(2,1,i,j+1);
               lhsc(2,2,i,j,k,c) =  tmp2 * fjac(2,2,i,j+1)
                    - tmp1 * njac(2,2,i,j+1)
                    - tmp1 * dy2;
               lhsc(2,3,i,j,k,c) =  tmp2 * fjac(2,3,i,j+1)
                    - tmp1 * njac(2,3,i,j+1);
               lhsc(2,4,i,j,k,c) =  tmp2 * fjac(2,4,i,j+1)
                    - tmp1 * njac(2,4,i,j+1);
               lhsc(2,5,i,j,k,c) =  tmp2 * fjac(2,5,i,j+1)
                    - tmp1 * njac(2,5,i,j+1);

               lhsc(3,1,i,j,k,c) =  tmp2 * fjac(3,1,i,j+1)
                    - tmp1 * njac(3,1,i,j+1);
               lhsc(3,2,i,j,k,c) =  tmp2 * fjac(3,2,i,j+1)
                    - tmp1 * njac(3,2,i,j+1);
               lhsc(3,3,i,j,k,c) =  tmp2 * fjac(3,3,i,j+1)
                    - tmp1 * njac(3,3,i,j+1)
                    - tmp1 * dy3;
               lhsc(3,4,i,j,k,c) =  tmp2 * fjac(3,4,i,j+1)
                    - tmp1 * njac(3,4,i,j+1);
               lhsc(3,5,i,j,k,c) =  tmp2 * fjac(3,5,i,j+1)
                    - tmp1 * njac(3,5,i,j+1);

               lhsc(4,1,i,j,k,c) =  tmp2 * fjac(4,1,i,j+1)
                    - tmp1 * njac(4,1,i,j+1);
               lhsc(4,2,i,j,k,c) =  tmp2 * fjac(4,2,i,j+1)
                    - tmp1 * njac(4,2,i,j+1);
               lhsc(4,3,i,j,k,c) =  tmp2 * fjac(4,3,i,j+1)
                    - tmp1 * njac(4,3,i,j+1);
               lhsc(4,4,i,j,k,c) =  tmp2 * fjac(4,4,i,j+1)
                    - tmp1 * njac(4,4,i,j+1)
                    - tmp1 * dy4;
               lhsc(4,5,i,j,k,c) =  tmp2 * fjac(4,5,i,j+1)
                    - tmp1 * njac(4,5,i,j+1);

               lhsc(5,1,i,j,k,c) =  tmp2 * fjac(5,1,i,j+1)
                    - tmp1 * njac(5,1,i,j+1);
               lhsc(5,2,i,j,k,c) =  tmp2 * fjac(5,2,i,j+1)
                    - tmp1 * njac(5,2,i,j+1);
               lhsc(5,3,i,j,k,c) =  tmp2 * fjac(5,3,i,j+1)
                    - tmp1 * njac(5,3,i,j+1);
               lhsc(5,4,i,j,k,c) =  tmp2 * fjac(5,4,i,j+1)
                    - tmp1 * njac(5,4,i,j+1);
               lhsc(5,5,i,j,k,c) =  tmp2 * fjac(5,5,i,j+1)
                    - tmp1 * njac(5,5,i,j+1)
                    - tmp1 * dy5;

            }
         }


//---------------------------------------------------------------------
//     outer most do loops - sweeping in i direction
//---------------------------------------------------------------------
         if (first == 1) {

//---------------------------------------------------------------------
//     multiply c(i,jstart,k) by b_inverse and copy back to c
//     multiply rhs(jstart) by b_inverse(jstart) and copy to rhs
//---------------------------------------------------------------------
//dir$ ivdep
            for (i = start(1,c); i <= isize; i++) {
               binvcrhs( &lhsb(1,1,i,jstart),
                              &lhsc(1,1,i,jstart,k,c),
                              &rhs(1,i,jstart,k,c) );
            }

         }

//---------------------------------------------------------------------
//     begin inner most do loop
//     do all the elements of the cell unless last 
//---------------------------------------------------------------------
         for (j = jstart+first; j <= jsize-last; j++) {
//dir$ ivdep
            for (i = start(1,c); i <= isize; i++) {

//---------------------------------------------------------------------
//     subtract A*lhs_vector(j-1) from lhs_vector(j)
//     
//     rhs(j) = rhs(j) - A*rhs(j-1)
//---------------------------------------------------------------------
               matvec_sub(&lhsa(1,1,i,j),
                               &rhs(1,i,j-1,k,c),&rhs(1,i,j,k,c));

//---------------------------------------------------------------------
//     B(j) = B(j) - C(j-1)*A(j)
//---------------------------------------------------------------------
               matmul_sub(&lhsa(1,1,i,j),
                               &lhsc(1,1,i,j-1,k,c),
                               &lhsb(1,1,i,j));

//---------------------------------------------------------------------
//     multiply c(i,j,k) by b_inverse and copy back to c
//     multiply rhs(i,1,k) by b_inverse(i,1,k) and copy to rhs
//---------------------------------------------------------------------
               binvcrhs( &lhsb(1,1,i,j),
                              &lhsc(1,1,i,j,k,c),
                              &rhs(1,i,j,k,c) );

            }
         }

//---------------------------------------------------------------------
//     Now finish up special cases for last cell
//---------------------------------------------------------------------
         if (last == 1) {

//dir$ ivdep
            for (i = start(1,c); i <= isize; i++) {
//---------------------------------------------------------------------
//     rhs(jsize) = rhs(jsize) - A*rhs(jsize-1)
//---------------------------------------------------------------------
               matvec_sub(&lhsa(1,1,i,jsize),
                               &rhs(1,i,jsize-1,k,c),&rhs(1,i,jsize,k,c));

//---------------------------------------------------------------------
//     B(jsize) = B(jsize) - C(jsize-1)*A(jsize)
//     call matmul_sub(aa,i,jsize,k,c,
//     $              cc,i,jsize-1,k,c,bb,i,jsize,k,c)
//---------------------------------------------------------------------
               matmul_sub(&lhsa(1,1,i,jsize),
                               &lhsc(1,1,i,jsize-1,k,c),
                               &lhsb(1,1,i,jsize));

//---------------------------------------------------------------------
//     multiply rhs(jsize) by b_inverse(jsize) and copy to rhs
//---------------------------------------------------------------------
               binvrhs( &lhsb(1,1,i,jsize),
                             &rhs(1,i,jsize,k,c) );
            }

         }
      }


      return;
}
#undef WORK_LHS_VEC
      


