
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <stdio.h>
#include "applu.incl"
#include "upcnpb.h"

void init_comm(int *argc, char ***argv) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//
//   initialize UPC and establish rank and size
//
// This is a module in the UPC implementation of LUSSOR
// pseudo application from the NAS Parallel Benchmarks. 
//
//---------------------------------------------------------------------


      extern int nodedim();
      int i, bsize1, bsize2, bsize3, bsize0;

//---------------------------------------------------------------------
//   initialize UPC parameters
//---------------------------------------------------------------------

      node = MYTHREAD;
      no_nodes = THREADS;
      if (no_nodes != NNODES_COMPILED) {
         if (node == 0)
            printf("Error: no_nodes not matched with compiled\n");
         upc_global_exit(1);
      }

#ifdef USE_BARRIER
      if (node == 0)
         printf(" Starting the BARRIER version\n");
#endif

//---------------------------------------------------------------------
//   allocate shared memory buffers
//---------------------------------------------------------------------
      if (no_nodes > 1) {
#ifdef USE_BARRIER
      	 bsize1 = 20*ISIZ2;
      	 bsize2 = 40*ISIZ2*ISIZ3;
      	 bsize3 = 2*(ISIZ1 + ISIZ2+2 + ISIZ3);
#else
      	 bsize1 =  5*(ISIZ1 + ISIZ2);
      	 bsize2 = 20*(ISIZ1 + ISIZ2) * ISIZ3;
      	 bsize3 = DMAX(2*(ISIZ1 + ISIZ2+2), ISIZ3);
#endif
	 bsize0 = bsize1 + bsize2 + bsize3;
//
	 u_buffer = (shared sh_buf_t *)upc_all_alloc(no_nodes, sizeof(sh_buf_t));
	 comm_count = (shared int *)upc_all_alloc(no_nodes, sizeof(int));
//
	 uptr(1) = 1;
	 vptr(1) = uptr(1) + bsize1;
	 wptr(1) = vptr(1) + bsize2;
//
#ifdef USE_BARRIER
      	 bsize1 = bsize1/4;
      	 bsize2 = bsize2/4;
      	 uptr(2) = uptr(1) + bsize1;
      	 uptr(3) = uptr(2) + bsize1;
      	 uptr(4) = uptr(3) + bsize1;
      	 vptr(2) = vptr(1) + bsize2;
      	 vptr(3) = vptr(2) + bsize2;
      	 vptr(4) = vptr(3) + bsize2;
      	 wptr(2) = wptr(1) + 2*ISIZ1;
      	 wptr(3) = wptr(2) + 2*(ISIZ2+2);
      	 wptr(4) = wptr(3) + ISIZ3;
#else
      	 bsize1 =  5*ISIZ2;
      	 bsize2 = 10*ISIZ2*ISIZ3;
      	 bsize3 = 10*ISIZ1*ISIZ3;
      	 uptr(2) = uptr(1) + bsize1;
      	 vptr(2) = vptr(1) + bsize2;
      	 vptr(3) = vptr(2) + bsize2;
      	 vptr(4) = vptr(3) + bsize3;
      	 wptr(2) = wptr(1) + 2*ISIZ1;
#endif
      }

//---------------------------------------------------------------------
//   initialize process id
//---------------------------------------------------------------------
      id = node;
      root = 0;

//---------------------------------------------------------------------
//   establish the size of the global group
//---------------------------------------------------------------------
      num = no_nodes;
      ndim   = nodedim(num);

      if (! convertdouble) {
         dp_type = 1;
      } else {
         dp_type = 0;
      }


      return;
}
