//-------------------------------------------------------------------------!
//                                                                         !
//        N  A  S     P A R A L L E L     B E N C H M A R K S  3.3         !
//                                                                         !
//                                   L U                                   !
//                                                                         !
//-------------------------------------------------------------------------!
//                                                                         !
//    This benchmark is part of the NAS Parallel Benchmark 3.3 suite.      !
//    It is described in NAS Technical Reports 95-020 and 02-007           !
//                                                                         !
//    Permission to use, copy, distribute and modify this software         !
//    for any purpose with or without fee is hereby granted.  We           !
//    request, however, that all derived work reference the NAS            !
//    Parallel Benchmarks 3.3. This software is provided "as is"           !
//    without express or implied warranty.                                 !
//                                                                         !
//    Information on NPB 3.3, including the technical report, the          !
//    original specifications, source code, results and information        !
//    on how to submit new results, is available at:                       !
//                                                                         !
//           http://www.nas.nasa.gov/Software/NPB/                         !
//                                                                         !
//    Send comments or suggestions to  npb@nas.nasa.gov                    !
//                                                                         !
//          NAS Parallel Benchmarks Group                                  !
//          NASA Ames Research Center                                      !
//          Mail Stop: T27A-1                                              !
//          Moffett Field, CA   94035-1000                                 !
//                                                                         !
//          E-mail:  npb@nas.nasa.gov                                      !
//          Fax:     (650) 604-3957                                        !
//                                                                         !
//-------------------------------------------------------------------------!

//---------------------------------------------------------------------
//
// Authors: S. Weeratunga
//          V. Venkatakrishnan
//          E. Barszcz
//          M. Yarrow
//          H. Jin  (UPC version)
//
//---------------------------------------------------------------------
#define G_MAIN
#include "applu.incl"
#include "upcnpb.h"

//---------------------------------------------------------------------
//    program applu
//---------------------------------------------------------------------
int main(int argc, char *argv[]) {

  unsigned int mymask = 1 << (MYTHREAD%16);
  //sched_setaffinity(0, sizeof(unsigned int), &mymask);
//---------------------------------------------------------------------
//
//   driver for the performance evaluation of the solver for
//   five coupled parabolic/elliptic partial differential equations.
//
//---------------------------------------------------------------------

      char class;
      int verified;
      double mflops, n3;
      int ierr;

      int i;
      for(i=2; i<=16; i++)
        timer_clear(i);

//---------------------------------------------------------------------
//   initialize communications
//---------------------------------------------------------------------
      init_comm(&argc, &argv);

//---------------------------------------------------------------------
//   read input data
//---------------------------------------------------------------------
      read_input();

//---------------------------------------------------------------------
//   set up processor grid
//---------------------------------------------------------------------
      proc_grid();

//---------------------------------------------------------------------
//   determine the neighbors
//---------------------------------------------------------------------
      neighbors();

//---------------------------------------------------------------------
//   set up sub-domain sizes
//---------------------------------------------------------------------
      subdomain();

//---------------------------------------------------------------------
//   set up coefficients
//---------------------------------------------------------------------
      setcoeff();

//---------------------------------------------------------------------
//   set the boundary values for dependent variables
//---------------------------------------------------------------------
      setbv();

//---------------------------------------------------------------------
//   set the initial values for dependent variables
//---------------------------------------------------------------------
      setiv();

//---------------------------------------------------------------------
//   compute the forcing term based on prescribed exact solution
//---------------------------------------------------------------------
      erhs();

//---------------------------------------------------------------------
//   perform one SSOR iteration to touch all data and program pages 
//---------------------------------------------------------------------
      ssor(1);

//---------------------------------------------------------------------
//   reset the boundary and initial values
//---------------------------------------------------------------------
      setbv();
      setiv();

//---------------------------------------------------------------------
//   perform the SSOR iterations
//---------------------------------------------------------------------
      ssor(itmax);

//---------------------------------------------------------------------
//   compute the solution error
//---------------------------------------------------------------------
      error();

//---------------------------------------------------------------------
//   compute the surface integral
//---------------------------------------------------------------------
      pintgr();

//---------------------------------------------------------------------
//   verification test
//---------------------------------------------------------------------
      if (id==0) {
         verify ( rsdnm, errnm, frc, &class, &verified );
         n3 = (double)( nx0+ny0+nz0 )/3.;
         mflops = (double)(itmax)*(1984.77*(double)( nx0 )
              *(double)( ny0 )
              *(double)( nz0 )
              -10923.3* n3*n3 
              +27770.9* n3
              -144010.)
              / (maxtime*1000000.);

         c_print_results("LU", class, nx0,
           ny0, nz0, itmax, /*NNODES_COMPILED,*/
           num, maxtime, mflops, "          floating point", verified, 
           NPBVERSION, COMPILETIME, NPB_CS1, NPB_CS2, NPB_CS3, NPB_CS4, NPB_CS5, NPB_CS6, 
           "(none)");
      }

      upc_global_exit(0);

      return 0;
}


