//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <stdio.h>
#include "applu.incl"
#include "upcnpb.h"

void ssor(int niter) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
//   to perform pseudo-time stepping SSOR iterations
//   for five nonlinear pde's.
//---------------------------------------------------------------------


//---------------------------------------------------------------------
//  local variables
//---------------------------------------------------------------------
      int i, j, k, m;
      int istep;
      int isiz1 = ISIZ1, isiz2 = ISIZ2, isiz3 = ISIZ3;
      double  tmp;
      double  delunm[5], tv[5*(ISIZ1)*(ISIZ2)];
#define delunm(i) delunm[i-1]
#define tv(i,j,k) tv[(i-1)+(5)*((j-1)+(ISIZ1)*(k-1))]

      double wtime;

#ifdef USE_BARRIER
      int kk, kst, kend, kkend, kkoff;


      kst = 2;
      kend = nz - 1;
      kkend = kend + xdim + ydim - 2;
      kkoff = row + col - 2;
#endif
      root = 0;
 
//---------------------------------------------------------------------
//   begin pseudo-time stepping iterations
//---------------------------------------------------------------------
      tmp = 1.0e+00 / ( omega * ( 2.0e+00 - omega ) ) ;

//---------------------------------------------------------------------
//   initialize a,b,c,d to zero (guarantees that page tables have been
//   formed, if applicable on given architecture, before timestepping).
//---------------------------------------------------------------------
      for (m = 1; m <= isiz2; m++) {
         for (k = 1; k <= isiz1; k++) {
            for (j = 1; j <= 5; j++) {
               for (i = 1; i <= 5; i++) {
                  a(i,j,k,m) = 0.e0;
                  b(i,j,k,m) = 0.e0;
                  c(i,j,k,m) = 0.e0;
                  d(i,j,k,m) = 0.e0;
               }
            }
         }
      }

//---------------------------------------------------------------------
//   compute the steady-state residuals
//---------------------------------------------------------------------
      rhs();
 
//---------------------------------------------------------------------
//   compute the L2 norms of newton iteration residuals
//---------------------------------------------------------------------
      l2norm( isiz1, isiz2, isiz3, nx0, ny0, nz0,
                   ist, iend, jst, jend,
                   rsd, rsdnm );
  
      upc_barrier;
#ifndef USE_BARRIER
      sync_init();
#endif
 
      timer_clear(1);
      timer_start(1);

//---------------------------------------------------------------------
//   the timestep loop
//---------------------------------------------------------------------
      for (istep = 1; istep <= niter; istep++) {

         if (id == 0) {
            if ( ( istep % 20) == 0 ||
                  istep == itmax ||
                  istep == 1) {
               if (niter > 1) printf(" Time step %4d\n", istep);
            }
         }
 
//---------------------------------------------------------------------
//   perform SSOR iteration
//---------------------------------------------------------------------
         for (k = 2; k <= nz - 1; k++) {
            for (j = jst; j <= jend; j++) {
               for (i = ist; i <= iend; i++) {
                  for (m = 1; m <= 5; m++) {
                     rsd(m,i,j,k) = dt * rsd(m,i,j,k);
                  }
               }
            }
         }
 
#ifndef USE_BARRIER
         for (k = 2; k <= nz -1; k++) {
#else
         bar_odd = 1;
         for (kk = 2; kk <= kkend; kk++) {
            bar_odd = 1 - bar_odd;
            k = kk - kkoff;
            if (k < kst || k > kend) {
               upc_barrier;
            }
            else {
#endif
//---------------------------------------------------------------------
//   form the lower triangular part of the jacobian matrix
//---------------------------------------------------------------------
            jacld(k);
 
//---------------------------------------------------------------------
//   perform the lower triangular solution
//---------------------------------------------------------------------
            blts( isiz1, isiz2, isiz3,
                       nx, ny, nz, k,
                       omega,
                       rsd,
                       a, b, c, d,
                       ist, iend, jst, jend, 
                       nx0, ny0, ipt, jpt);
#ifdef USE_BARRIER
            }
#endif
         }

#ifndef USE_BARRIER
         for (k = nz - 1; k >= 2; k--) {
#else
         for (kk = kkend; kk >= 2; kk--) {
            bar_odd = 1 - bar_odd;
            k = kk - kkoff;
            if (k < kst || k > kend) {
               upc_barrier;
            }
            else {
#endif
//---------------------------------------------------------------------
//   form the strictly upper triangular part of the jacobian matrix
//---------------------------------------------------------------------
            jacu(k);

//---------------------------------------------------------------------
//   perform the upper triangular solution
//---------------------------------------------------------------------
            buts( isiz1, isiz2, isiz3,
                       nx, ny, nz, k,
                       omega,
                       rsd, tv,
                       d, a, b, c,
                       ist, iend, jst, jend,
                       nx0, ny0, ipt, jpt);
#ifdef USE_BARRIER
            }
#endif
         }
 
//---------------------------------------------------------------------
//   update the variables
//---------------------------------------------------------------------
 
         for (k = 2; k <= nz-1; k++) {
            for (j = jst; j <= jend; j++) {
               for (i = ist; i <= iend; i++) {
                  for (m = 1; m <= 5; m++) {
                     u( m, i, j, k ) = u( m, i, j, k )
                          + tmp * rsd( m, i, j, k );
                  }
               }
            }
         }
 
//---------------------------------------------------------------------
//   compute the max-norms of newton iteration corrections
//---------------------------------------------------------------------
         if ( ( istep % inorm ) == 0 ) {
            l2norm( isiz1, isiz2, isiz3, nx0, ny0, nz0,
                         ist, iend, jst, jend,
                         rsd, delunm );
//            if ( ipr == 1 && id == 0 ) then
//                write (*,1006) ( delunm(m), m = 1, 5 )
//            else if ( ipr == 2 && id == 0 ) then
//                write (*,'(i5,f15.6)') istep,delunm(5)
//            end if
         }
 
//---------------------------------------------------------------------
//   compute the steady-state residuals
//---------------------------------------------------------------------
         rhs();
 
//---------------------------------------------------------------------
//   compute the max-norms of newton iteration residuals
//---------------------------------------------------------------------
         if ( ( ( istep % inorm ) == 0 ) ||
              ( istep == itmax ) ) {
            l2norm( isiz1, isiz2, isiz3, nx0, ny0, nz0,
                         ist, iend, jst, jend,
                         rsd, rsdnm );
//            if ( ipr == 1&&id==0 ) then
//                write (*,1007) ( rsdnm(m), m = 1, 5 )
//            end if
         }

//---------------------------------------------------------------------
//   check the newton-iteration residuals against the tolerance levels
//---------------------------------------------------------------------
         if ( ( rsdnm(1) < tolrsd(1) ) &&
              ( rsdnm(2) < tolrsd(2) ) &&
              ( rsdnm(3) < tolrsd(3) ) &&
              ( rsdnm(4) < tolrsd(4) ) &&
              ( rsdnm(5) < tolrsd(5) ) ) {
//            if (ipr == 1 && id==0) then
//               write (*,1004) istep
//            end if
            return;
         }
 
      }
 
      timer_stop(1);
      wtime = timer_read(1);
 

      mupc_reduce_r8( &wtime, &maxtime, 1, 
                      SMP_OPR_MAX, root );
 


      return;
       
#undef delunm
#undef tv
}
