//---------------------------------------------------------------------
//---------------------------------------------------------------------
//
//    Point-to-point synchronization routines
//
//       H. Jin  <hjin@nas.nasa.gov>
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <time.h>
#include <sched.h>
#include "applu.incl"
#include "upcnpb.h"


#define _MAXCOUNT 1000
#define _SMP_WAIT(cond) \
   {  if (cond) { \
      	 volatile int c = 0, c2 = 0; \
	 while (cond) { \
	    upc_fence; \
	    if (++c > _MAXCOUNT) { \
	       sched_yield(); c = 0; \
	       if (++c2 > 1000) { \
                  struct timespec tr={0,1000};\
                  nanosleep(&tr, (struct timespec *)0);\
	       } \
	    } \
	 } \
      } \
   }

typedef struct sync_buf_s {
   int v[NNODES_XDIM][2];
} sync_buf_t;

shared sync_buf_t sync_ptr[THREADS];

static int *sptr, sid;

void sync_init(void)
{
   int i;

   if (no_nodes < 2) return;

   for (i = row-1; i <= row+1; i++)
      if (i >= 1 && i <= xdim)
         sync_ptr[node].v[i-1][1] = 0;
   for (i = col-1; i <= col+1; i++)
      if (i >= 1 && i <= ydim)
         sync_ptr[node].v[i-1][0] = 0;
   upc_barrier;
}

void sync_testwait(int nid, int dir, int sdir)
{
   sid = ((dir == 0)? row : col) + sdir - 1;

   sptr = (int *)&sync_ptr[node].v[sid][dir];
   _SMP_WAIT (*sptr == 0);
}

void sync_ackwait(int nid, int dir)
{
   *sptr = 0;
   sync_ptr[nid].v[sid][dir] = 0;
   upc_fence;
}

void sync_testnotify(int nid, int dir, int sdir)
{
   sid = ((dir == 0)? row : col) + sdir - 1;

   sptr = (int *)&sync_ptr[node].v[sid][dir];
   _SMP_WAIT (*sptr == 1);
}

void sync_setnotify(int nid, int dir)
{
   *sptr = 1;
   sync_ptr[nid].v[sid][dir] = 1;
   upc_fence;
}
