
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include "header.h"
#include "upcnpb.h"

void copy_faces() {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// this function copies the face values of a variable defined on a set 
// of cells to the overlap locations of the adjacent sets of cells. 
// Because a set of cells interfaces in each direction with exactly one 
// other set, we only need to fill six different buffers. We could try to
// overlap communication with computation, by computing
// some internal values while communicating boundary values, but this
// adds so much overhead that it's not clearly useful. 
//---------------------------------------------------------------------

       int i, j, k, c, m, p0, p1, 
               p2, p3, p4, p5, b_size[6], ss[6], 
               sr[6], sr_n[6];
#define b_size(i) b_size[i]
#define ss(i) ss[i]
#define sr(i) sr[i]
#define sr_n(m) sr_n[m]

//---------------------------------------------------------------------
//      exit immediately if there are no faces to be copied           
//---------------------------------------------------------------------
       if (no_nodes == 1) {
          compute_rhs();
          return;
       }


       ss(0) = start_send_east;
       ss(1) = start_send_west;
       ss(2) = start_send_north;
       ss(3) = start_send_south;
       ss(4) = start_send_top;
       ss(5) = start_send_bottom;

       sr(0) = start_recv_east;
       sr(1) = start_recv_west;
       sr(2) = start_recv_north;
       sr(3) = start_recv_south;
       sr(4) = start_recv_top;
       sr(5) = start_recv_bottom;

       sr_n(0) =   successor(1);
       sr_n(1) = predecessor(1);
       sr_n(2) =   successor(2);
       sr_n(3) = predecessor(2);
       sr_n(4) =   successor(3);
       sr_n(5) = predecessor(3);

       b_size(0) = east_size   ;
       b_size(1) = west_size   ;
       b_size(2) = north_size  ;
       b_size(3) = south_size  ;
       b_size(4) = top_size    ;
       b_size(5) = bottom_size ;

//---------------------------------------------------------------------
// because the difference stencil for the diagonalized scheme is 
// orthogonal, we do not have to perform the staged copying of faces, 
// but can send all face information simultaneously to the neighboring 
// cells in all directions          
//---------------------------------------------------------------------
       TIMER_START(T_BPACK);
       p0 = 0;
       p1 = 0;
       p2 = 0;
       p3 = 0;
       p4 = 0;
       p5 = 0;

       for (c = 1; c <= ncells; c++) {
          for (m = 1; m <= 5; m++) {

//---------------------------------------------------------------------
//            fill the buffer to be sent to eastern neighbors (i-dir)
//---------------------------------------------------------------------
             if (cell_coord(1,c) != ncells) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = cell_size(1,c)-2; i <= cell_size(1,c)-1; i++) {
                         ul_buffer(ss(0)+p0) = u(i,j,k,m,c);
                         p0 = p0 + 1;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
//            fill the buffer to be sent to western neighbors 
//---------------------------------------------------------------------
             if (cell_coord(1,c) != 1) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = 0; i <= 1; i++) {
                         ul_buffer(ss(1)+p1) = u(i,j,k,m,c);
                         p1 = p1 + 1;
                      }
                   }
                }


             }

//---------------------------------------------------------------------
//            fill the buffer to be sent to northern neighbors (j_dir)
//---------------------------------------------------------------------
             if (cell_coord(2,c) != ncells) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = cell_size(2,c)-2; j <= cell_size(2,c)-1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         ul_buffer(ss(2)+p2) = u(i,j,k,m,c);
                         p2 = p2 + 1;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
//            fill the buffer to be sent to southern neighbors 
//---------------------------------------------------------------------
             if (cell_coord(2,c)!= 1) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= 1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         ul_buffer(ss(3)+p3) = u(i,j,k,m,c);
                         p3 = p3 + 1;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
//            fill the buffer to be sent to top neighbors (k-dir)
//---------------------------------------------------------------------
             if (cell_coord(3,c) != ncells) {
                for (k = cell_size(3,c)-2; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         ul_buffer(ss(4)+p4) = u(i,j,k,m,c);
                         p4 = p4 + 1;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
//            fill the buffer to be sent to bottom neighbors
//---------------------------------------------------------------------
             if (cell_coord(3,c)!= 1) {
                 for (k = 0; k <= 1; k++) {
                    for (j = 0; j <= cell_size(2,c)-1; j++) {
                       for (i = 0; i <= cell_size(1,c)-1; i++) {
                          ul_buffer(ss(5)+p5) = u(i,j,k,m,c);
                          p5 = p5 + 1;
                       }
                    }
                 }
              }

//---------------------------------------------------------------------
//          m loop
//---------------------------------------------------------------------
           }

//---------------------------------------------------------------------
//       cell loop
//---------------------------------------------------------------------
        }
       TIMER_STOP(T_BPACK);


// .... synchronize all updates
      TIMER_START(T_EXCH);
      for (c = 0; c < 6; c++) {
         upc_memput(&u_buffer(node,ss(c)-1), &ul_buffer(ss(c)),
                    sizeof(double)*b_size(c));
      }
      upc_barrier;
      for (c = 0; c < 6; c++) {
         upc_memget(&ul_buffer(sr(c)), &u_buffer(sr_n(c),sr(c)-1),
                    sizeof(double)*b_size(c));
      }
      TIMER_STOP(T_EXCH);

//---------------------------------------------------------------------
// unpack the data that has just been received;             
//---------------------------------------------------------------------
       TIMER_START(T_BPACK);
       p0 = 0;
       p1 = 0;
       p2 = 0;
       p3 = 0;
       p4 = 0;
       p5 = 0;

       for (c = 1; c <= ncells; c++) {
          for (m = 1; m <= 5; m++) {

             if (cell_coord(1,c) != 1) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = -2; i <= -1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(1)+p0);
                         p0 = p0 + 1;
                      }
                   }
                }
             }

             if (cell_coord(1,c) != ncells) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = cell_size(1,c); i <= cell_size(1,c)+1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(0)+p1);
                         p1 = p1 + 1;
                      }
                   }
                }
             }
 
             if (cell_coord(2,c) != 1) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = -2; j <= -1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(3)+p2);
                         p2 = p2 + 1;
                      }
                   }
                }

             }
 
             if (cell_coord(2,c) != ncells) {
                for (k = 0; k <= cell_size(3,c)-1; k++) {
                   for (j = cell_size(2,c); j <= cell_size(2,c)+1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(2)+p3);
                         p3 = p3 + 1;
                      }
                   }
                }
             }

             if (cell_coord(3,c) != 1) {
                for (k = -2; k <= -1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(5)+p4);
                         p4 = p4 + 1;
                      }
                   }
                }
             }

             if (cell_coord(3,c) != ncells) {
                for (k = cell_size(3,c); k <= cell_size(3,c)+1; k++) {
                   for (j = 0; j <= cell_size(2,c)-1; j++) {
                      for (i = 0; i <= cell_size(1,c)-1; i++) {
                         u(i,j,k,m,c) = ul_buffer(sr(4)+p5);
                         p5 = p5 + 1;
                      }
                   }
                }
             }

//---------------------------------------------------------------------
//         m loop            
//---------------------------------------------------------------------
          }

//---------------------------------------------------------------------
//      cells loop
//---------------------------------------------------------------------
       }
       TIMER_STOP(T_BPACK);

//---------------------------------------------------------------------
// now that we have all the data, compute the rhs
//---------------------------------------------------------------------
       compute_rhs();

       return;
#undef b_size
#undef ss
#undef sr
#undef ss_g
#undef sr_g
}
