//---------------------------------------------------------------------
//---------------------------------------------------------------------
//
//    Point-to-point synchronization routines
//
//       H. Jin  <hjin@nas.nasa.gov>
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <time.h>
#include <sched.h>
#include "header.h"
#include "upcnpb.h"


#define _MAXCOUNT 1000
#define _SMP_WAIT(cond) \
   {  if (cond) { \
      	 volatile int c = 0, c2 = 0; \
	 while (cond) { \
	    upc_fence; \
	    if (++c > _MAXCOUNT) { \
	       sched_yield(); c = 0; \
	       if (++c2 > 1000) { \
                  struct timespec tr={0,1000};\
                  nanosleep(&tr, (struct timespec *)0);\
	       } \
	    } \
	 } \
      } \
   }

typedef struct sync_buf_s {
   int v[MAXCELLS*MAXCELLS];
} sync_buf_t;

shared sync_buf_t sync_ptr[THREADS];

static int *sptr;


void sync_init(void)
{
   int i;

   if (no_nodes < 2) return;

   for (i = 0; i < no_nodes; i++)
      sync_ptr[node].v[i] = 0;
   upc_barrier;
}

void sync_testwait(int nid)
{
   if (no_nodes < 2) return;

   sptr = (int *)&sync_ptr[node].v[nid];
   _SMP_WAIT (*sptr == 0);
}

void sync_ackwait(int nid)
{
   if (no_nodes < 2) return;

   *sptr = 0;
   sync_ptr[nid].v[nid] = 0;
   upc_fence;
}

void sync_testnotify(int nid)
{
   if (no_nodes < 2) return;

   sptr = (int *)&sync_ptr[node].v[node];
   _SMP_WAIT (*sptr == 1);
}

void sync_setnotify(int nid)
{
   if (no_nodes < 2) return;

   *sptr = 1;
   sync_ptr[nid].v[node] = 1;
   upc_fence;
}
