

subroutine caf_queue_shifter(dim,zeon,zeon0,memax,safety_extra_bounds,left_rank_MPI, right_rank_MPI,&
                             mpi_rank,recvQueue,recvQueue0,buffer_size,&
                             buffer_chunks,empty_RQ_rate, send_buffer, send_buffer0, send_buffer_counter,&
                             sb_chunks_counter, nr_repeats,repeat_cnt,&
                             get_dest,queue_it,recv_divisor,my_shift_neighbors,&
                             nproc_partd,nproc_toroidal,myrank_partd,myrank_toroidal)
  use queue_lock_module
  use array_of_PointerToMatrix_module
 
  implicit none
  
  include 'mpif.h' ! just for using the MPI's wallclock timer
  
  integer :: time_levels, left_rank_MPI, right_rank_MPI,zeon_it,recv_divisor,mpi_rank
  integer :: world_rank, world_size, left_rank, right_rank, amount_rem,dim,time_levels
  integer :: safety_extra_bounds, all_recv_buffers_empty, empty_RQ_rate,nr_repeats
  integer :: thread_1_rank, thread_0_rank, temp, repeat_cnt
  integer :: memax, m, i, j, dest, all_world_ranks_are_done, mtop, lasth
  integer :: send_zeon_cnt,local_queue_position,printzeon
  integer :: omp_get_num_threads, omp_get_thread_num, nthreads, thread_id, send_buffer_dest
  integer :: nr_of_holes_can_be_filled, already_filled_holes, temp_send_zeon_cnt
  integer :: nr_particles_added_to_end, buffer_chunks
  integer :: nproc_partd,nproc_toroidal,myrank_partd,myrank_toroidal,rank_partd,rank_toroidal
  integer :: recv_queue_length, nr_particles_fill_holes, nr_particles_add_to_end


  logical :: got_lock, got_lock2, test
  REAL(8),INTENT(INOUT),ALLOCATABLE,DIMENSION(:,:),CODIMENSION[:,:] :: recvQueue,recvQueue0

  integer(8),allocatable :: queue_it(:)[:,:]
  integer(8) buffer_size, dest_queue_position, queue_bound, send_length, help, offset
  ! kzeon stores zeon position of each sent particle
  integer,allocatable :: kzeon(:),my_shift_neighbors(:) 
  real(8),allocatable :: send_buffer(:,:,:),send_buffer0(:,:,:)
  integer,allocatable :: send_buffer_counter(:),get_dest(:),sb_chunks_counter(:)
  
  real(8),INTENT(INOUT),ALLOCATABLE,DIMENSION(:,:) :: zeon,zeon0

  !sync all
  allocate( kzeon(1:(memax) ) )
    
  world_rank = mpi_rank + 1 ! THIS_IMAGE()
  world_size = nproc_toroidal ! NUM_IMAGES()
  left_rank =  left_rank_MPI+1
  right_rank = right_rank_MPI+1
  rank_partd = myrank_partd + 1
  rank_toroidal = myrank_toroidal

  if(repeat_cnt==1 .AND. world_rank==1) write(*,*) ' '
  if(repeat_cnt==1 .AND. world_rank==1) write(*,*) '    --------------------- '
  
  if(repeat_cnt==1 .AND. world_rank==1) then
    write(*,*) '  -> CAF Queue Benchmark starts: world_rank=', world_rank, ' within world_size=', world_size, &
    ' nr_repeats=',nr_repeats, 'memax=', memax, 'buffer_size=', buffer_size, 'buffer_chunks=', buffer_chunks,&
    'left_rank=',left_rank,'right_rank=',right_rank,'rank_partd=',rank_partd,'rank_toroidal=',rank_toroidal,&
    'my_shift_neighbors=',my_shift_neighbors
    write(*,*) ' '
  endif  

  
  thread_0_rank = 1
  thread_1_rank = 2
  dest = -1
  send_buffer_counter = 0
  sb_chunks_counter = 0
  nr_of_holes_can_be_filled = 0
  already_filled_holes = 0
  send_zeon_cnt = 0
  temp_send_zeon_cnt = 0
  all_world_ranks_are_done = 1
  all_recv_buffers_empty = 0
  nr_particles_added_to_end = 0
  printzeon = 1
  test = .false.
  queue_it(1) = 1

  sync images([my_shift_neighbors]) ! need to sync; especially because of resetting the queue iterator
 
  !iterate through the zeon array and fill up the send buffer
    zeon_it=0
    do
 10   zeon_it=zeon_it+1
      if(zeon_it > memax) exit

      dest = int(zeon(1,zeon_it))
  
      if( (dest >= 0) .AND. (dest .NE. rank_toroidal) ) then ! means that this particle will be shifted

        kzeon(send_zeon_cnt + 1) = zeon_it ! mark the holes
        send_zeon_cnt = send_zeon_cnt + 1
      
        ! if the send buffer for image dest is NOT full: fill it
        send_buffer_counter(dest) = send_buffer_counter(dest) + 1
        send_buffer_dest = send_buffer_counter(dest)
     
        send_buffer(1:dim, send_buffer_dest+(sb_chunks_counter(dest)*buffer_size), get_dest(dest)) = zeon(:,zeon_it)
        send_buffer(dim+1:2*dim, send_buffer_dest+(sb_chunks_counter(dest)*buffer_size), get_dest(dest)) = zeon0(:,zeon_it)

        ! empty recv queue; but rarely to avoid too many locks
        if( mod(send_zeon_cnt,empty_RQ_rate) == 0) then ! send_buffer_dest == 10*buffer_size

!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
!  ++++++++++++++++++++ FIRING UP MESSAGES / EMPTY RECV Q +++++++++++++++++++++  

          !lock(queue_lock[world_rank])         
          do ! try to get my "own" lock to empty my recv Q. if do not get it. jump over.
            lock( queue_lock[rank_partd,rank_toroidal], acquired_lock=got_lock2)
            if (got_lock2) exit
            goto 20
          end do 
          
          sync memory

          !local_queue_position = queue_it(1) ! can read from [1 .. loc-1] (from loc can be filled)
          help = 0
          local_queue_position = amo_aaddf(queue_it(1),help)

          ! nr particles which can be stored: send_zeon_cnt - already_filled_holes
          queue_bound = max(1,local_queue_position-(send_zeon_cnt-already_filled_holes)+1)
          
          
          !$omp parallel do private(m) & 
          !$omp shared(already_filled_holes,local_queue_position,queue_bound,kzeon,zeon,zeon0,recvQueue)
          do m=(local_queue_position-1),queue_bound,-1
            zeon(:,kzeon(already_filled_holes+m-queue_bound+1) ) = recvQueue(1:dim,m)
            zeon0(:,kzeon(already_filled_holes+m-queue_bound+1) ) = recvQueue(dim+1:2*dim,m)
          enddo
          already_filled_holes = already_filled_holes + ((local_queue_position-1)-queue_bound ) + 1
          
          !queue_it(1) = queue_bound
          call amo_aadd(queue_it(1),queue_bound-(local_queue_position))

          sync memory
          unlock(queue_lock[rank_partd,rank_toroidal])
        
        endif
! #########################################################################################################   
 20     if( send_buffer_dest == buffer_size ) then  
        
          !lock(queue_lock[dest])
          do
            lock( queue_lock[rank_partd,dest], acquired_lock=got_lock)
            if (got_lock) exit
            if( sb_chunks_counter(dest) < (buffer_chunks-1) ) then ! only if sendbuffer is not totally full         
              sb_chunks_counter(dest) = sb_chunks_counter(dest) + 1
              send_buffer_counter(dest) = 0
              goto 10
            endif
          end do        
  
          sync memory
          
          !FULL: fire up a message to dest (fill its recv queue)

          !The amo_aaddf function (atomic add and fetch) adds value to the var object and returns the new value.
          help = 0
          dest_queue_position = amo_aaddf(queue_it(1)[rank_partd,dest],help)

          offset = sb_chunks_counter(dest)*buffer_size + send_buffer_counter(dest)

          if( dest_queue_position+offset-1 > memax/recv_divisor ) then
            !if(world_rank==thread_0_rank) 
            write(*,*) '!!! RecvQ, world_rank=',world_rank, &
            'send > memax // dest_queue_position+offset-1=',&
            dest_queue_position+offset-1,'memax/recv_divisor=',memax/recv_divisor
            stop
          endif

          recvQueue(:,(dest_queue_position):(dest_queue_position+offset-1) )[rank_partd,dest]=&
            send_buffer(:,1:offset,get_dest(dest))

          !queue_it(1)[rank_partd,dest] = dest_queue_position+buffer_size ! now move the queue iterator to last position
          call amo_aadd(queue_it(1)[rank_partd,dest],offset)    

          !R1 sync memory
          unlock(queue_lock[rank_partd,dest])
          sync memory
          
          send_buffer_counter(dest) = 0 ! so, send buffer for this destination is totally empty now
          sb_chunks_counter(dest) = 0
         
!  ++++++++++++++++++++ FIRING UP MESSAGES / EMPTY RECV Q +++++++++++++++++++++        
!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++         

        endif
      endif
    enddo
    

    





    ! send remaining send buffers which are not fully filled
    do i = 1, world_size
      send_length = sb_chunks_counter(i)*buffer_size + send_buffer_counter(i)
      if( send_length > 0 ) then ! in case there is something to send; but only write into it when it is empty

!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
!  +++++++++++++ FIRING UP REMAINING MESSAGES / EMPTY RECV Q ++++++++++++++++++      
        lock(queue_lock[rank_partd,i])
        sync memory
     
        !dest_queue_position = queue_it(1)[rank_partd,i] ! get the position in the remote recv queue where to insert
        help = 0
        dest_queue_position = amo_aaddf(queue_it(1)[rank_partd,i],help)          

        recvQueue(:,(dest_queue_position):(dest_queue_position+send_length-1) )[rank_partd,i] =&
          send_buffer(:,1:send_length,get_dest(i))

        !queue_it(1)[rank_partd,i] = dest_queue_position+send_length ! now move the queue iterator to last position
        call amo_aadd(queue_it(1)[rank_partd,i],send_length)

        sync memory
        unlock(queue_lock[rank_partd,i])

!  +++++++++++++ FIRING UP REMAINING MESSAGES / EMPTY RECV Q ++++++++++++++++++         
!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++          
   
      endif
    enddo
 
    !sync all ! get sure that all images have emptied their send buffers and local queues are ready to be read
    sync images([my_shift_neighbors])

    ! now every image can fill up its remaining particles from the recv queue
    
    ! nr particles have to be read from recv queue and added to zeon: "queue_it(1)-1"
    recv_queue_length = queue_it(1)-1
    
    ! nr particles which can be overwritten in zeon: "send_zeon_cnt - already_filled_holes"

    ! nr of particles can be taken from recvQueue to fill overwritable zeon particles:
    !   min( "queue_it(1)-1", "send_zeon_cnt - already_filled_holes" )
    nr_particles_fill_holes = min( recv_queue_length, send_zeon_cnt - already_filled_holes )
    
    ! nr of particles taken from recvQueue which will be placed at the end of zeon array:
    !   "queue_it(1)-1" - min( "queue_it(1)-1", "send_zeon_cnt - already_filled_holes" )
    nr_particles_add_to_end = recv_queue_length - nr_particles_fill_holes
    
  
    !$omp parallel do private(m) & 
    !$omp shared(already_filled_holes,nr_particles_fill_holes,queue_bound,kzeon,zeon,zeon0,recvQueue)
    do m=1, nr_particles_fill_holes
      zeon( :, kzeon(already_filled_holes+m) ) = recvQueue(1:dim, m)
      zeon0( :, kzeon(already_filled_holes+m) ) = recvQueue(dim+1:2*dim,m)
    enddo

    already_filled_holes = already_filled_holes + nr_particles_fill_holes
    
    ! add particles to end of zeon OR
    if( nr_particles_add_to_end > 0 ) then
      zeon( :, memax+1 : memax+nr_particles_add_to_end ) = &
        recvQueue(1:dim, recv_queue_length-nr_particles_add_to_end+1 : recv_queue_length)
      zeon0( :, memax+1 : memax+nr_particles_add_to_end ) = &
        recvQueue(dim+1:2*dim, recv_queue_length-nr_particles_add_to_end+1 : recv_queue_length)  
        
      memax=memax+nr_particles_add_to_end
    ! fill remaining holes: it might happen that local zeon array had more holes than received particles  
    elseif( nr_particles_add_to_end <= 0 .AND. already_filled_holes < send_zeon_cnt ) then
      ! new algorithm taken from GTS shifter :   
      mtop=memax
      memax=memax+already_filled_holes-send_zeon_cnt
      ! fill the hole
      lasth=send_zeon_cnt
                
      do i = (already_filled_holes+1), send_zeon_cnt !for unfilled holes
        m=kzeon(i)
        if(m>memax) exit  !Break out of the DO loop
        do while(mtop == kzeon(lasth))
          mtop=mtop-1
          lasth=lasth-1
        enddo
      
        zeon(:,m) = zeon(:,mtop)
        zeon0(:,m) = zeon0(:,mtop)
        mtop=mtop-1
      enddo
    endif   

  repeat_cnt = repeat_cnt + 1
  
  !sync all
  deallocate( kzeon )

end subroutine caf_queue_shifter  
