! ########################################################################
! Author: Robert Preissl, NERSC - LBNL
! September 2010
! Benchmark for shifting particles in a torus
!  (simulates shfter PIC step in the GTS application from PPPL Princeton)
! ########################################################################




program GTS

  use array_of_PointerToMatrix_module
  use destination_comp
  use queue_lock_module

  implicit none
    
  INTERFACE
    subroutine caf_queue_shifter(dim,zeon2,zeon20,memax2, safety_extra_bounds,left_rank_MPI,right_rank_MPI,&
                                 mpi_rank,recvQueue2,recvQueue20,buffer_size,&
                                 buffer_chunks,empty_RQ_rate, send_buffer2, send_buffer20,send_buffer_counter2,&
                                 sb_chunks_counter2, nr_repeats,repeat_cnt,&
                                 get_dest,queue_it2,recv_divisor,my_shift_neighbors,&
                                 nproc_partd,nproc_toroidal,myrank_partd,myrank_toroidal)
                                 
      use array_of_PointerToMatrix_module
                                 
      integer,intent(IN) :: dim, safety_extra_bounds, left_rank_MPI, right_rank_MPI,recv_divisor
      integer :: buffer_chunks,empty_RQ_rate,nr_repeats,repeat_cnt,time_levels,mpi_rank
      integer :: nproc_partd,nproc_toroidal,myrank_partd,myrank_toroidal
      integer,intent(INOUT) :: memax2
      integer(8) :: buffer_size
      real(8),intent(INOUT),allocatable,dimension(:,:) :: recvQueue2[:,:],recvQueue20[:,:]
      real(8),intent(INOUT),allocatable,dimension(:,:) :: zeon2,zeon20
      real(8),allocatable,dimension(:,:,:) :: send_buffer2,send_buffer20
      integer,allocatable :: send_buffer_counter2(:),sb_chunks_counter2(:)
      integer,allocatable :: get_dest(:)
      integer,allocatable :: my_shift_neighbors(:)
      integer(8),allocatable :: queue_it2(:)[:,:]
    end subroutine caf_queue_shifter                      
  END INTERFACE  
  
  
  
  
  include 'mpif.h'
  
  integer :: memax, i, j, safety_extra_bounds,particle_domain_location,toroidal_domain_location
  integer :: partd_comm,toroidal_comm,nproc_partd,myrank_partd,nproc_toroidal,myrank_toroidal,ntoroidal
  integer :: ierr,rc,left_rank_toroidal,right_rank_toroidal,memax2,memax3,dim,COPY_memax3,numtasks,mpi_rank
  integer :: buffer_chunks,empty_RQ_rate, COPY_memax2, nr_repeats, repeat_cnt, time_levels,upper_bound
  integer :: COPY_memax, buffer_size, memax4, COPY_memax4, memax5, COPY_memax5,recv_left_divisor,nr_threads
  integer :: nr_of_possible_send_destinations,lower_bound,temp_image_rank,m, autotune,toroidal_image_rank
  integer(8) :: buffer_size2
  integer :: omp_get_num_threads, omp_get_thread_num
  integer,allocatable,dimension(:) :: get_dest,send_mask,my_global_toroidal_neighbors_CAF
  integer,allocatable,dimension(:,:) :: get_recv
  integer, parameter :: mpi_Rsize=MPI_DOUBLE_PRECISION


! arrays for CAF 1D recvQ benchmark  
  real(8),allocatable,dimension(:,:) :: zeon2,zeon20    ! the zeon array for the 2nd caf benchmark
  real(8),allocatable,dimension(:) :: COPY_zeon2,COPY2_zeon2    ! copy of zeon array; to execute benchmark several
  real(8),allocatable,dimension(:,:) :: recvQueue2[:,:],recvQueue20[:,:] !  several times. (zeon will be modified)
  real(8),allocatable,dimension(:,:,:) :: send_buffer2,send_buffer20
  integer,allocatable,dimension(:) :: send_buffer_counter2
  integer,allocatable,dimension(:) :: sb_chunks_counter2
  integer,allocatable :: my_shift_neighbors(:)
  integer(8),allocatable :: queue_it2(:)[:,:]
  character(len=10) ic_recvQ(13)

! arrays for MPI GTS shifter benchmark 
  real(8),allocatable,dimension(:,:) :: mpi_zeon,mpi_zeon0 ! the zeon array for the mpi benchmark
  real(8),allocatable,dimension(:) :: COPY_mpi_zeon,COPY2_mpi_zeon
  integer,allocatable :: kzeon_mpi(:),iright_mpi(:),ileft_mpi(:) ! stores zeon position of each sent particle
  
! arrays for CAF-MPI-GTS shifter benchmark 
  real(8),allocatable,dimension(:,:) :: CAFmpi_zeon,CAFmpi_zeon0 ! the zeon array for the caf-mpiAlg benchmark
  real(8),allocatable,dimension(:) :: COPY_CAFmpi_zeon,COPY2_CAFmpi_zeon
  real(8),allocatable,dimension(:,:) ::  recvleft_CM[:,:], recvleft0_CM[:,:]  ! coarrays have to have same size in Cray Ftn, so 
  real(8),allocatable,dimension(:,:) :: recvright_CM[:,:],recvright0_CM[:,:] !  allocate one before. size = memax/4


  ! Initialize the MPI library:
  call MPI_INIT(ierr)
  if (ierr .ne. MPI_SUCCESS) then
     print *,'Error starting MPI program. Terminating.'
     call MPI_ABORT(MPI_COMM_WORLD, rc, ierr)
  end if
  call MPI_COMM_SIZE(MPI_COMM_WORLD,numtasks,ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD,mpi_rank,ierr)
  
  nr_repeats = 100 ! how often benchmark is executed within a loop
  repeat_cnt = 1
  dim        = 7 
  memax      = 1000000 ! length of zeon array
!  npartdom   = 4  ! number of PEs per toroidal domain  ---> is defined in modules.F90

  ! values of zeon array, which are a multiple of send_mask("x") are shifted
  nr_of_possible_send_destinations = 10
  recv_left_divisor = 10
  allocate( send_mask(1:nr_of_possible_send_destinations) )
 !-------------- to the LEFT --------------
  send_mask(1)=82421! shift to "myrank-5"
  send_mask(2)=2467 ! shift to "myrank-4"
  send_mask(3)=1249 ! shift to "myrank-3"
  send_mask(4)=149  !                etc.
  send_mask(5)=11
 !-------------- to the RIGHT --------------
  send_mask(6)=82531! shift to "myrank+5"
  send_mask(7)=2503 ! shift to "myrank+4"
  send_mask(8)=1259 ! shift to "myrank+3"
  send_mask(9)=131  !                etc.
  send_mask(10)=13
  
  ! numtasks must be a multiple of npartdom so change npartdom accordingly
  if( mod(numtasks,npartdom) .NE. 0 ) then 
    write(*,*) 'please change npartdom'
    STOP
  endif
  ntoroidal=numtasks/npartdom
  allocate( my_global_toroidal_neighbors_CAF(1:ntoroidal-1) )
  
  ! numtasks must be a multiple of npartdom so change npartdom accordingly
  if( ntoroidal < nr_of_possible_send_destinations ) then 
    write(*,*) 'ntoroidal too small --- ntoroidal=', ntoroidal
    STOP
  endif

  
  ! do the MPI 1D domain decomposition in toroidal direction according to GTS
  call set_particle_decomp(mpi_rank,numtasks,npartdom,ntoroidal,particle_domain_location,toroidal_domain_location,&
                           partd_comm,toroidal_comm,nproc_partd,myrank_partd,nproc_toroidal,myrank_toroidal,&
                           left_rank_toroidal,right_rank_toroidal,my_global_toroidal_neighbors_CAF)
                           
  ! Get the toroidal rank of the CAF image 
  toroidal_image_rank = myrank_toroidal + 1 ! myrank_toroidal is the MPI toroidal rank                      
                            
  if( nr_of_possible_send_destinations >= nproc_toroidal ) then
    write(*,*) 'nr_of_possible_send_destinations >= nproc_toroidal'
    STOP
  endif    

  allocate( get_dest(1:nproc_toroidal) ) ! mapping of destination value to actual storage position
  allocate( get_recv(1:nr_of_possible_send_destinations,1:nproc_toroidal) ) ! mapping of recv value to storage position

  
  memax2 = memax ! length of zeon array will change during the shifter benchmark.
  memax3 = memax
  memax4 = memax
  memax5 = memax
  
  safety_extra_bounds = 1000 ! zeon array can have more than memax particles at the end of shift

  !$OMP PARALLEL
  nr_threads = omp_get_num_threads()
  !$OMP END PARALLEL



  
  
  
! ######################### CAF-1D RecvQ Benchmark #########################

  ! ++++++++++++++++ SEND BUFFER PARAMETER ++++++++++++++++
  buffer_size2 = 5300
  buffer_chunks = 1 ! means that the send buffer has "buffer_chunks" pieces of length "buffer_size2"
  empty_RQ_rate = 10000000 ! means that every "empty_RQ_rate" send elements, the recv queue is emptied
  ! +++++++++++++++++++++++++++++++++++++++++++++++++++++++
  if(buffer_chunks > 1) then
    write(*,*) 'ERROR buffer_chunks > 1; :operator does not work any more in nolock-shifter'
    stop
  endif  

  allocate(  recvQueue2(1:(1*2*dim),1:(memax2/(recv_left_divisor/2)))[nproc_partd,*] )
  allocate( queue_it2(1:2)[nproc_partd,*] )
  
  allocate( zeon2(1:dim,1:(memax2+safety_extra_bounds)), zeon20(1:dim,1:(memax2+safety_extra_bounds)) )
  allocate( COPY_zeon2(1:(memax2+safety_extra_bounds)) )
  allocate( COPY2_zeon2(1:(memax2+safety_extra_bounds)) )
  allocate( send_buffer2(1:(2*dim), 1:(buffer_size2*buffer_chunks), 1:nr_of_possible_send_destinations) )
  allocate( send_buffer_counter2(1:nproc_toroidal) )
  allocate( sb_chunks_counter2(1:nproc_toroidal) )
  allocate( my_shift_neighbors(1:(nr_of_possible_send_destinations) ) )
  
  ! compute a mapping function: destination el.of(1..nproc_toroidal) -> (1..nr_of_possible_send_destinations)
  upper_bound = mod(toroidal_image_rank+(nr_of_possible_send_destinations/2)-1,nproc_toroidal) + 1
  lower_bound = mod(toroidal_image_rank-(nr_of_possible_send_destinations/2)-1+nproc_toroidal,nproc_toroidal) + 1
  get_dest = -1
  j = 1
  do i = 1,nproc_toroidal
    if( ( (lower_bound < upper_bound) .AND. ((i >= lower_bound) .AND. (i<= upper_bound)) ) .OR. &
        ( (lower_bound > upper_bound) .AND. (((i >= 1) .AND. (i<= upper_bound)) .OR. &
          ((i >= lower_bound) .AND. (i<= nproc_toroidal))) ) ) then
      if( i.NE.toroidal_image_rank ) then
        get_dest(i) = j ! mapping function: 1..nproc_toroidal -> 1..nr_of_possible_send_destinations
        my_shift_neighbors(j)=(i-1)*nproc_partd+myrank_partd+1 ! +1 because of MPI - CAF conversion // globally computed
        j = j + 1
      endif  
    endif
  end do
  
  repeat_cnt = 1


  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  
  do i=1,dim
    do j=1,memax2
      zeon2(i,j) = j + mpi_rank
      zeon20(i,j) = j + mpi_rank
    enddo
  enddo
  do i=1,dim
    do j=memax2+1,memax2+safety_extra_bounds
      zeon2(i,j) = -100.0
      zeon20(i,j) = -100.0
    enddo
  enddo
  
  call compute_toroidal_dest(0,zeon2(1,1:memax2),toroidal_image_rank,nproc_toroidal,memax2,safety_extra_bounds,&
                             nr_of_possible_send_destinations,send_mask,1,myrank_partd)
  
  COPY_zeon2 = zeon2(1,:)
  COPY2_zeon2 = zeon2(2,:)
  COPY_memax2 = memax2
  
  do i=1, nr_repeats
  
! *************** CAF single recv queued Particle shifter ***************
    call caf_queue_shifter(dim,zeon2, zeon20, memax2, safety_extra_bounds,left_rank_toroidal,&
                           right_rank_toroidal,mpi_rank,&
                           recvQueue2,recvQueue20,buffer_size2,buffer_chunks,empty_RQ_rate, send_buffer2,&
                           send_buffer20,send_buffer_counter2,sb_chunks_counter2, nr_repeats,repeat_cnt,&
                           get_dest,queue_it2,(recv_left_divisor/2),my_shift_neighbors,nproc_partd,nproc_toroidal,&
                           myrank_partd,toroidal_image_rank)
! *********************************************************************** 
   
    zeon2(1,:) = COPY_zeon2
    zeon2(2,:) = COPY2_zeon2
    memax2 = COPY_memax2
    recvQueue2 = 0
    
  
  enddo
  
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
                             
  deallocate(zeon2,zeon20)
  deallocate(COPY_zeon2,COPY2_zeon2)
  deallocate(recvQueue2)
  deallocate(send_buffer_counter2)
  deallocate(sb_chunks_counter2)
  deallocate(send_buffer2)
  deallocate(queue_it2)
  deallocate( my_shift_neighbors )
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
! ########################################################################  
  
  

  
  
  
  
! ######################### MPI GTS Shifter Benchmark #########################  
  allocate( mpi_zeon(1:dim,1:(memax3+safety_extra_bounds)),mpi_zeon0(1:dim,1:(memax3+safety_extra_bounds)))
  allocate( COPY_mpi_zeon(1:(memax3+safety_extra_bounds)) )
  allocate( COPY2_mpi_zeon(1:(memax3+safety_extra_bounds)) )
  allocate( kzeon_mpi(1:memax3),iright_mpi(1:memax3),ileft_mpi(1:memax3) )
  do i=1,dim
    do j=1,memax3
       mpi_zeon(i,j) = j + mpi_rank
       mpi_zeon0(i,j) = j + mpi_rank
    enddo
  enddo
  do i=1,dim
    do j=memax3+1,memax3+safety_extra_bounds
      mpi_zeon(i,j) = -100.0
      mpi_zeon0(i,j) = -100.0
    enddo
  enddo 
  
  call compute_toroidal_dest(1,mpi_zeon(1,1:memax3),toroidal_image_rank,nproc_toroidal,memax3,safety_extra_bounds,&
                             nr_of_possible_send_destinations,send_mask,0,myrank_partd)

  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  
  COPY_mpi_zeon = mpi_zeon(1,:)
  COPY2_mpi_zeon = mpi_zeon(2,:)
  COPY_memax3 = memax3
 
  repeat_cnt = 1
  
  do i=1, nr_repeats
  
! *************** MPI-OpenMP Particle shifter ***************
    call mpi_openmp_shifter(dim,mpi_zeon, mpi_zeon0, memax3, safety_extra_bounds,&
                            myrank_toroidal,left_rank_toroidal,right_rank_toroidal, nr_repeats,repeat_cnt,&
                            kzeon_mpi,iright_mpi,ileft_mpi,toroidal_comm,nproc_toroidal,myrank_partd)
! *********************************************************** 

    mpi_zeon(1,:) = COPY_mpi_zeon
    mpi_zeon(2,:) = COPY2_mpi_zeon
    memax3 = COPY_memax3
  
  enddo
  
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  deallocate(mpi_zeon,mpi_zeon0,COPY_mpi_zeon,COPY2_mpi_zeon)
  deallocate(kzeon_mpi,iright_mpi,ileft_mpi)
 
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
! ########################################################################  
  
  
  
  
  deallocate(get_dest,get_recv,send_mask,my_global_toroidal_neighbors_CAF)
  
  call MPI_FINALIZE(ierr)
 
end program GTS












! -> test run on Franklin. 24 cores. -N4, -d1. using standard locking queue shifter
!
!*******************************************************
!   Using npartdom= 2  and ntoroidal= 12
! *******************************************************
! 
! mype= 0   nproc_toroidal= 12  myrank_toroidal= 0   nproc_partd= 2  myrank_partd= 0
! Module.F90 : CAF_rank= 1 send_mask_counter= 12,  405,  800,  6697,  89392,  12,  399,  794,  7564,  68760 sum_to_the_left= 97306 sum
! _to_the_right= 77529
!  
!     --------------------- 
!   -> CAF Queue Benchmark starts: world_rank= 1  within world_size= 12  nr_repeats= 100 memax= 1000000 buffer_size= 5300 buffer_chunk
! s= 1 left_rank= 12 right_rank= 2 rank_partd= 1 rank_toroidal= 1 my_shift_neighbors= 3,  5,  7,  9,  11,  15,  17,  19,  21,  23
!  
! CAF 1D RECV-Q WALL TIME (in SEC):
! alloc     shifter   1st       read-QI   fill-Q    crit-1    2nd       3rd       CRIT      writ-QI   fillSB    lockAQ    SYNC      
!  1.43E-03  2.38E+01  1.32E+01  0.00E+00  0.00E+00  0.00E+00  9.78E-01  9.61E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  7.55E-04  2.38E+01  1.29E+01  0.00E+00  0.00E+00  0.00E+00  1.23E+00  9.61E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  1.43E-03  2.38E+01  1.32E+01  0.00E+00  0.00E+00  0.00E+00  1.45E+00  9.78E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main queue CAF :  mpi_rank= 0 MPI timing=, 23.814532518386841  CPU timing= 23.481466000000001 Sysclock= 23.81302892149877 memax2=
! 1000000
!  
!     --------------------- 
!  
!   -> MPI Benchmark starts: myrank_toroidal= 0 myrank_partd= 0  left_rank_MPI= 11  right_rank_MPI= 1 memax= 1000000
!  
!  
! MPI WALL TIME (in SEC):
! 1st       holes     Send/R    unpack    allred    part.W    pack                
!  2.84E+01  7.55E+00  5.04E+00  2.56E+00  8.34E-01  3.54E+00  7.89E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  2.84E+01  7.63E+00  5.07E+00  2.59E+00  8.37E-01  3.51E+00  7.80E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  2.84E+01  7.70E+00  5.42E+00  2.71E+00  1.03E+00  3.61E+00  7.90E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main MPI :  mpi_rank= 0 MPI timing=, 28.360428333282471  CPU timing= 26.789676000000014 memax3= 1000000
!  
!     --------------------- 
!  
!   -> CAF-MPIAlg Benchmark starts: world_rank= 1 rank_partd= 1 rank_toroidal= 1 world_size (nproc_toroidal)= 12  left_rank= 12  right
! _rank= 2 memax= 1000000 nproc_partd= 2
!  
!  
! CAF MPI-Alg SHIFTER WALL TIME (in SEC):
! 1st       holes     Send/R    unpack    allred    part.W                        
!  3.19E+01  8.51E+00  7.69E+00  2.93E+00  1.98E+00  2.94E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  3.19E+01  8.54E+00  7.72E+00  2.98E+00  1.95E+00  2.89E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  3.19E+01  8.56E+00  7.95E+00  3.05E+00  2.11E+00  2.94E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main CAF-MPIAlg :  mpi_rank= 0 MPI timing=, 32.093572616577148  CPU timing= 31.577972000000031 memax5= 1000000
!Application 4459832 resources: utime 0, stime 0






! -> test run on Franklin. 24 cores. -N2, -d2. using non-locking queue shifter
! *******************************************************
!   Using npartdom= 2  and ntoroidal= 12
! *******************************************************
! 
! mype= 0   nproc_toroidal= 12  myrank_toroidal= 0   nproc_partd= 2  myrank_partd= 0
! Module.F90 : CAF_rank= 1 send_mask_counter= 12,  405,  800,  6697,  89392,  12,  399,  794,  7564,  68760 sum_to_the_left= 97306 sum
! _to_the_right= 77529
!  
!     --------------------- 
!   -> CAF Queue Benchmark starts: world_rank= 1  within world_size= 12  nr_repeats= 100 memax= 1000000 buffer_size= 5300 buffer_chunk
! s= 1 left_rank= 12 right_rank= 2 rank_partd= 1 rank_toroidal= 1 my_shift_neighbors= 3,  5,  7,  9,  11,  15,  17,  19,  21,  23
!  
!  
! CAF 1D RECV-Q WALL TIME (in SEC):
! alloc     shifter   1st       read-QI   fill-Q    crit-1    2nd       3rd       CRIT      writ-QI   fillSB    lockAQ    SYNC      
!  6.04E-04  1.21E+01  7.13E+00  0.00E+00  0.00E+00  0.00E+00  5.31E-01  4.42E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  5.67E-04  1.21E+01  7.03E+00  0.00E+00  0.00E+00  0.00E+00  5.72E-01  4.47E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  6.21E-04  1.21E+01  7.13E+00  0.00E+00  0.00E+00  0.00E+00  6.18E-01  4.56E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main queue CAF :  mpi_rank= 0 MPI timing=, 12.088536262512207  CPU timing= 24.069502000000007 Sysclock= 12.087590824903829 memax2=
! 1000000
!  
!     --------------------- 
!  
!   -> MPI Benchmark starts: myrank_toroidal= 0 myrank_partd= 0  left_rank_MPI= 11  right_rank_MPI= 1 memax= 1000000
!  
!  
! MPI WALL TIME (in SEC):
! 1st       holes     Send/R    unpack    allred    part.W    pack                
!  1.56E+01  4.11E+00  3.99E+00  9.90E-01  6.25E-01  2.23E+00  3.15E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  1.56E+01  4.10E+00  4.03E+00  9.84E-01  6.00E-01  2.24E+00  3.15E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  1.57E+01  4.13E+00  4.15E+00  1.02E+00  8.06E-01  2.27E+00  3.15E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main MPI :  mpi_rank= 0 MPI timing=, 15.624157428741455  CPU timing= 30.993939000000019 memax3= 1000000
!  
!     --------------------- 
!  
!   -> CAF-MPIAlg Benchmark starts: world_rank= 1 rank_partd= 1 rank_toroidal= 1 world_size (nproc_toroidal)= 12  left_rank= 12  right
! _rank= 2 memax= 1000000 nproc_partd= 2
!  
!  
! CAF MPI-Alg SHIFTER WALL TIME (in SEC):
! 1st       holes     Send/R    unpack    allred    part.W                        
!  1.61E+01  4.27E+00  2.99E+00  1.61E+00  1.04E+00  2.14E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  1.61E+01  4.28E+00  2.97E+00  1.62E+00  1.07E+00  2.13E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
!  1.61E+01  4.29E+00  3.10E+00  1.63E+00  1.15E+00  2.14E+00  0.00E+00  0.00E+00
!
!  0.00E+00  0.00E+00  0.00E+00  0.00E+00  0.00E+00
! X main CAF-MPIAlg :  mpi_rank= 0 MPI timing=, 16.264243841171265  CPU timing= 32.06600600000003 memax5= 1000000