#include <stdio.h>
#include <stdlib.h>
#include <upc.h>
//
// A is addressed in the same manner as in the serial case
// 
typedef struct {
  shared [] float *row;
  size_t row_size;
} Tri_matrix;

shared [1] size_t n, i_row;
shared [1] Tri_matrix *A;

int main(void) {
  size_t i, i_local, j, n_elem;
  int next_thread;
  float *row;

  if (MYTHREAD == 0) scanf("%li %li",(size_t *) &n,(size_t *) &i_row);
// casts to (size_t *) are required and work because data are located
// on executing thread
  upc_barrier;
// n and i_row are now accessible from each thread

  A = (shared [1] Tri_matrix *) 
       upc_all_alloc(n, sizeof(Tri_matrix));
//
// create triangular matrix
  n_elem = 0;
  for (i=MYTHREAD; i<n; i+=THREADS) {
    A[i].row_size = n - i;
    A[i].row = 
         (shared [] float *) upc_alloc(A[i].row_size*sizeof(float));
    n_elem += n - i;
  }
//
// initialize matrix A(i, j) = i + j
  for (i=MYTHREAD; i<n; i+=THREADS) {
//  cast to local pointer for performance
    row = (float *) A[i].row;
    for (j=0; j<n-i; j++) {
      row[j] = (float) (i + 1) + (float) (j + 1);
    }
  }
//
// write (one-based) row number i_row to standard output
  upc_barrier;
  printf(" Row %li on thread %i:", i_row, MYTHREAD);
  for (j=0; j<n-i_row+1; j++) {
    printf(" %4.1f", A[i_row-1].row[j]);
  }
  printf("\n");
  printf("Number of elements on thread %i: %li\n", 
          MYTHREAD,n_elem);
//
// clean up
  upc_barrier; 
  for (i=MYTHREAD; i<n; i+=THREADS) {
    upc_free(A[i].row);
  }
  if (MYTHREAD == 0) upc_free(A);
}
